/* 
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 */

/*=========================================================================
 * KVM
 *=========================================================================
 * SYSTEM:    KVM
 * SUBSYSTEM: Java-level debugger
 * FILE:      debugger.c
 * OVERVIEW:  This file defines the Java-level debugger interface for
 *            KVM based on JPDA (Java Platform Debug Architecture).
 * AUTHOR:    Jonathan Coles, Sun Labs (summer 2000)
 *            Bill Pittore, Sun Labs
 *=======================================================================*/

/*=========================================================================
 * Include files
 *=======================================================================*/

#include <global.h>

#if ENABLE_JAVA_DEBUGGER

#include <debuggerCommands.h>
#include <thread.h>
#ifdef PILOT
  #include <sys_types.h>
  #include <netinet_in.h>
#endif

#ifdef WINDOWS
  #define WIN32_LEAN_AND_MEAN
  #define NOMSG
  #ifdef GCC
    #include <winsock.h>
  #else
    #include <winsock2.h>
  #endif
#endif

/*========================================================================
 * Variables
 *========================================================================*/

static void **funcArray;
static void **KVMFuncArray;

static  PacketInputStream inputStream;
static  PacketOutputStream outputStream;
static bool_t resumeOK = FALSE;

HASHTABLE debuggerHashTable;

ID_HASH_ENTRY bucketFreeList = NULL;

POINTERLIST debugRoot = NULL;
unsigned long *bucketMap = NULL;
unsigned long bucketsAllocated;

cell **PermObjectRoot = NULL;
int PermObjectLength = 0;

unsigned long uniqueDebuggerID = 1;

VMEventPtr eventHead = NIL;
EVENTMODIFIER modHead = NIL;

VMEvent ssEvent;        /* only one step active at a time */

struct CE_Modifiers *CE_ModHead = NIL;

struct Modifiers *ModHead = NIL;

static long requestID = 1;

bool_t vmDebugReady;
bool_t debuggerActive;

bool_t waitOnSuspend;
bool_t allThreadsSuspended = FALSE;

short debuggerPort = DEFAULT_DEBUGGER_PORT;

long debuggerNotifyList=Dbg_EventKind_NONE;

unsigned long dummyLong;

#define CURRENTTHREAD (CurrentThread == NULL ? NULL : CurrentThread->javaThread)
#define GET_VMTHREAD(threadID) (getObjectPtr(threadID) == NULL ? NULL : ((JAVATHREAD)getObjectPtr(threadID))->VMthread)
/*========================================================================
 * Debugging operations
 *========================================================================*/

void nop(PacketInputStream *in, PacketOutputStream *out)
{
    outStream_setError(out, JDWP_Error_NOT_IMPLEMENTED);
    return;
}

static void * allocAlignedBytes(long size) {

    void *p;

#if COMPILER_SUPPORTS_LONG
    /*
     * mallocBytes allocates a cell sized block which means that
     * the address returned is always 4 byte aligned.  We need
     * 8 byte aligned so we add 4 more bytes to size then
     * re-adjust the address of newMod
     */
        p = callocPermanentObject(ByteSizeToCellSize((size + 4)));
    /* roundup to next 8 byte boundary */
    p = (void *)(((unsigned long)p + 7) & ~7);
#else
        p = callocPermanentObject(ByteSizeToCellSize(size));
#endif
    return (p);

}
bool_t 
isObjectTag(unsigned char tag) {
    return (tag == JDWP_Tag_OBJECT) ||
           (tag == JDWP_Tag_STRING) ||
           (tag == JDWP_Tag_THREAD) ||
           (tag == JDWP_Tag_THREAD_GROUP) ||
           (tag == JDWP_Tag_CLASS_LOADER) ||
           (tag == JDWP_Tag_CLASS_OBJECT) ||
           (tag == JDWP_Tag_ARRAY);
}

BYTE 
specificTypeKey(OBJECT object) 
{

    if (object == NIL) {
        return JDWP_Tag_OBJECT;
    } else if (isAssignableTo(object->ofClass, (CLASS)JavaLangString)) {
        return JDWP_Tag_STRING;
    } else if (isAssignableTo(object->ofClass, (CLASS)JavaLangClass)) {
        return JDWP_Tag_CLASS_OBJECT;
    } else {
        return (IS_ARRAY_CLASS(object->ofClass) ? JDWP_Tag_ARRAY :
            JDWP_Tag_OBJECT);
    }
}

static void setResumeOK(bool_t val) {
    resumeOK = val;
}

/*========================================================================
 * Function:        validThread()
 * Overview:        determines whether a thread actually exists
 *                  
 * Interface:
 *  parameters:     thread
 *  returns:        bool_t
 *
 * Notes:
 *=======================================================================*/

bool_t validThread(THREAD thread)
{
    THREAD tptr;

    tptr = AllThreads;

    if (tptr == NIL) return FALSE;
    if (tptr == thread) return TRUE;
    if (tptr == (THREAD)ONLY_THREADGROUP_ID) return FALSE;

    for (tptr=tptr->nextAliveThread; tptr && tptr != AllThreads; tptr = tptr->nextAliveThread) {
        if (tptr == thread)
            return TRUE;
    }
    return FALSE;
}

static bool_t validObject(OBJECT object)
{
    return (object != NIL) && (object->ofClass != NIL);
}

static unsigned long debuggerHash(void *p)
{
    return (((unsigned long)(((unsigned long)p >> 8) + 37)));
}

static void freeBucket(ID_HASH_ENTRY bucket) {

    ID_HASH_ENTRY tmpBucket, *freeBucketPtr, *lastBucketPtr;
    unsigned long index;

    index = (bucket->key >> DEBUGGER_INDEX_SHIFT) & DEBUGGER_INDEX_MASK;
    freeBucketPtr = (ID_HASH_ENTRY *)&debuggerHashTable->bucket[index];
    lastBucketPtr = freeBucketPtr;
    for (tmpBucket = *freeBucketPtr;
        tmpBucket != NULL; tmpBucket = tmpBucket->next) {
        if (bucket == tmpBucket) {
            *lastBucketPtr = tmpBucket->next;
            break;
         }
         lastBucketPtr = &tmpBucket->next;
    }
    bucket->next = bucketFreeList;
    bucketFreeList = bucket;
    bucket->key = 0;
}

/*========================================================================
 * Function:        addToDebugRoot()
 * Overview:        add this object to the debug root
 * Interface:
 *   parameters:    pointer to object
 *   returns:       index into array
 *
 * Notes:
 *                 The debugRoot contains object references and these
 *                 references will be updated by the GC since debugRoot is a
 *                 weak Pointerlist type.  We also keep a parallel array
 *                 called the bucketMap which contains a pointer to the
 *                 hash bucket that will contain this entry.  We do this
 *                 so we can easily re-use buckets since they are allocated
 *                 from permanent memory.  If we re-use a debugRoot entry
 *                 (because it was GC'd and NULL'd out) we set the bucket->key
 *                 value to 0 so that the getObjectID code will notice that
 *                 this bucket is free and re-use it.
 *                 Note that classes/objects that are in permanent memory
 *                 have an ID that is their actual address.  We distinguish
 *                 between these and transient objects by bit 0 being set in
 *                 the ID.
 *=======================================================================*/

static unsigned int addToDebugRoot(void *object, ID_HASH_ENTRY bucket)
{
    unsigned long i;
    unsigned long length = debugRoot->length;
    POINTERLIST tmpList;
    ID_HASH_ENTRY tmpBucket;
    unsigned long *tmpPtr;

    /*
     * Weak  pointerlists start at offset 1, offset 0 used by GC
     * 
     */

    for (i = 1; i < length -1; i++) {
        if (debugRoot->data[i].cellp == NULL) {
            debugRoot->data[i].cellp = (cell *)object;
            if (bucketMap[i] != 0) {
                /*
                 * This debugRoot entry was freed by the GC, so lets free 
                 * up the bucket that corresponds to it.
                 */
                tmpBucket = (ID_HASH_ENTRY)bucketMap[i];
#if INCLUDEDEBUGCODE
                if (i != tmpBucket->rootIndex)
                    fatalError("debugHashtable inconsistent");
#endif
                freeBucket(tmpBucket);
            }
            bucketMap[i] = (unsigned long)bucket;
            return(i);
        }
    }
    if (i == length - 1) {
        if (i >= DEBUGGER_INDEX_MASK)
            fatalError("Out of debugger root slots");

        /* we didn't find a free slot, put it at the end */
        debugRoot->data[i].cellp = (cell *)object;
        if (bucketMap[i] != 0) {
            /* see comment above */
            tmpBucket = (ID_HASH_ENTRY)bucketMap[i];
#if INCLUDEDEBUGCODE
            if (i != tmpBucket->rootIndex)
                fatalError("debugHashtable inconsistent");
#endif
            freeBucket(tmpBucket);
        }
        bucketMap[i] = (unsigned long)bucket;
        /* we're full, re-allocate */

        tmpList = (POINTERLIST)callocObject(SIZEOF_POINTERLIST(debugRoot->length + 20), GCT_WEAK_POINTERLIST);
        tmpList->length = length + 20;
        memcpy(tmpList->data, debugRoot->data, length << log2CELL);
        debugRoot = tmpList;

        tmpPtr = (unsigned long *)mallocBytes(debugRoot->length * CELL);
        memset(tmpPtr, 0, debugRoot->length * CELL);
        memcpy(tmpPtr, bucketMap, length << log2CELL);
        bucketMap = tmpPtr;
    }
    return (i);
}

static unsigned long getObjectIDInternal(void *object)
{
    HASHTABLE table;
    unsigned long index;
    ID_HASH_ENTRY bucket, *bucketPtr;

    table = debuggerHashTable;
    index = debuggerHash(object) % table->bucketCount;
    bucketPtr = (ID_HASH_ENTRY *)&table->bucket[index];
    for (bucket = *bucketPtr; bucket != NULL; bucket = bucket->next) {
        if (debugRoot->data[bucket->rootIndex].cellp == object)
            return (bucket->key);
    }
    return (0);
}

/*========================================================================
 * Function:        getObjectID()
 * Overview:        given an object, find/create it's ID
 * Interface:
 *   parameters:    pointer to object
 *   returns:       index into array
 *
 * Notes:
 *=======================================================================*/

unsigned long getObjectID(void *object)
{
    HASHTABLE table;
    unsigned long key;
    unsigned long index;
    ID_HASH_ENTRY bucket, *bucketPtr;

    if (!vmDebugReady)
        return (0);
    if (object == NULL)
        return (0);
    if (!inCurrentHeap(object))
        return (unsigned long)object;

    key = getObjectIDInternal(object);
    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(object, object);
    if (key == 0) {
        table = debuggerHashTable;
        index = debuggerHash(object) % table->bucketCount;
        bucketPtr = (ID_HASH_ENTRY *)&table->bucket[index];
        bucket = NULL;
        /* need to create a new one, first try to find an unused bucket */
        if (bucketFreeList != NULL) {
            bucket = bucketFreeList;
            bucketFreeList = bucket->next;
        }
        if (bucket == NULL) {
            bucket = (ID_HASH_ENTRY)callocPermanentObject(SIZEOF_ID_HASH_ENTRY);
            bucketsAllocated++;
        }
        bucket->rootIndex = addToDebugRoot(object, bucket);
        bucket->next = *bucketPtr;
        *bucketPtr = bucket;
        bucket->key = NEXT_UNIQUE_DEBUGGER_ID() + ((index & DEBUGGER_INDEX_MASK) << DEBUGGER_INDEX_SHIFT) + 1;
        key = bucket->key;

    }
    END_TEMPORARY_ROOTS
    return (key);
}

/*========================================================================
 * Function:        getObjectPtr()
 * Overview:        given an ID, find the object it maps to
 * Interface:
 *   parameters:    ID number
 *   returns:       void pointer
 *
 * Notes:
 *=======================================================================*/

static void * getObjectPtr(unsigned long key)
{
    unsigned long index;
    ID_HASH_ENTRY bucket, *bucketPtr;

    if (!vmDebugReady)
        return NULL;
    if ((key & 1) == 0)
        return ((void *)key);

    index = (key >> DEBUGGER_INDEX_SHIFT) & DEBUGGER_INDEX_MASK;
    bucketPtr = (ID_HASH_ENTRY *)&debuggerHashTable->bucket[index];
    for (bucket = *bucketPtr; bucket != NULL; bucket = bucket->next) {
        if (key == bucket->key)
            return (debugRoot->data[bucket->rootIndex].cellp);
    }
    return NULL;
}

static METHOD getMethodFromIndex(long classID, int index)
{
    INSTANCE_CLASS clazz = (INSTANCE_CLASS)getObjectPtr(classID);
    int i = 0;

    FOR_EACH_METHOD(thisMethod, clazz->methodTable)
        if (i == index) {
            return (thisMethod);
        }
        i++;
    END_FOR_EACH_METHOD
    return NULL;
}

/*========================================================================
 * Function:        getModifier()
 * Overview:        return a specific type of modifier contained within
 *                  an event
 * Interface:
 *  parameters:     event pointer, modifier type
 *  returns:        pointer to an event modifier struct
 *
 * Notes:
 *=======================================================================*/

static EVENTMODIFIER getModifier(VMEventPtr ep, BYTE type)
{
    EVENTMODIFIER mod;
    
    for (mod = ep->mods; mod != NIL; mod = mod->next) {
        if (mod->kind == type) {
            return mod;
        }
    }
    
    return NIL;
}

/*========================================================================
 * Function:        findBreakEvent()
 * Overview:        Search through our list of breakpoints to find this one 
 *              that we hit
 * Interface:
 *  parameters:     pointer to opcode, class, method , offset
 *  returns:        pointer to breakpoint structure
 *
 * Notes:
 *=======================================================================*/

static VMEventPtr findBreakEvent(ByteCode *opcode, unsigned long classID, unsigned long thisMethod, unsigned long offset)
{
    VMEventPtr ep = eventHead;
    EVENTMODIFIER mod;
    
    if (ep == NIL)
        return (NIL);       /* no breakpoints??? */

    do {
        if (ep->kind == JDWP_EventKind_BREAKPOINT) {
        if ((mod = getModifier(ep, JDWP_EventRequest_Set_Out_modifiers_Modifier_LocationOnly)) != NIL) {
            if (mod->u.loc.classID == classID && 
                mod->u.loc.methodIndex == thisMethod && 
                mod->u.loc.offset == offset     /*&&
                mod->u.loc.tag    == tag */) {

                if (opcode != NIL)
                    *opcode = mod->u.loc.opcode;

                return (ep); 

                    }
         }
        }
    } while ((ep = ep->next));
    return NIL;
}

#define SIMPLE_MATCH 1
#define BEGIN_MATCH 2
#define END_MATCH 3

static bool_t checkClassMatch(CEModPtr cep, EVENTMODIFIER em) {

    CLASS clazz = getObjectPtr(cep->loc.classID);
    char *modName;
    char *index;
    char *className = NULL; 
    char *packageName = NULL; 
    char *fullName;
    char *p;
    int i, type;
    unsigned int len = 0;

    START_TEMPORARY_ROOTS
    /* validate class name */
    if (clazz->baseName != NULL)
        className =  UStringInfo(clazz->baseName);

    /* validate package name */
    if (clazz->packageName != NULL)
        packageName = UStringInfo(clazz->packageName);

    if (className != NULL)
    len = strlen(className);
    if (packageName != NULL)
    len += strlen(packageName);
    IS_TEMPORARY_ROOT(fullName, mallocBytes(len + 2));
    if (packageName != NULL && *packageName != '\0') {
        strcpy(fullName, packageName);
        strcat(fullName, ".");
        strcat(fullName, className);
    } else {
    strcpy(fullName, className);
    }
    /* make sure that fullName is a 'standard' classname with '.'s */
    p = fullName;
    while (*p != '\0') {
        if (*p == '/' || *p == '\\')
            *p = '.';
        p++;
    }
    index = strchr((modName = em->u.classMatch.className), '*');
    if (index == NIL)
        type = SIMPLE_MATCH;
    else if (index == modName)
        type = BEGIN_MATCH;
    else if (index == modName + strlen(modName) - 1)
        type = END_MATCH;
    switch(type) {
    case SIMPLE_MATCH:
        if (strcmp(modName, fullName) == 0) {
        goto breakout;
        }
        break;

    case BEGIN_MATCH:
        /* start matching from the end */
        if (strlen(modName) > strlen(fullName) + 1)
            break;
        i = strlen(modName) - 1;
        while (i-- >= 0) {
            if (modName[i] == '*'){
            goto breakout;
            }
            if (modName[i] != fullName[i])
                break;
        }
        break;

    case END_MATCH:
        if (strlen(modName) > strlen(fullName) + 1)
            break;
        i = strlen(modName) - 1;
        if (strncmp(modName, fullName, i) == 0) {
        goto breakout;
        }
        break;
    }
    END_TEMPORARY_ROOTS
    return (FALSE);
breakout:
    return (TRUE);
}

/*========================================================================
 * Function:        satisfiesModifiers()
 * Overview:        determines whether an event which has occured
 *                  satisfies its modifiers
 * Interface:
 *  parameters:     event pointer, current event pointer, mod pointer
 *
 * Notes:
 *=======================================================================*/

static bool_t satisfiesModifiers(VMEventPtr ep, CEModPtr cep, EVENTMODIFIER em) 
{
    
    if (em == NIL)
        return TRUE;
    switch (em->kind) {
    case JDWP_EventRequest_Set_Out_modifiers_Modifier_Conditional:
        /* reserved for future use */
        break;
    case JDWP_EventRequest_Set_Out_modifiers_Modifier_ThreadOnly:
        if (cep->threadID != em->u.threadID)
            return FALSE;
        break;
    case JDWP_EventRequest_Set_Out_modifiers_Modifier_ClassOnly:
        if (em->u.classID == cep->loc.classID)
            return TRUE;
        else
            return FALSE;
    case JDWP_EventRequest_Set_Out_modifiers_Modifier_ClassMatch:
        if (checkClassMatch(cep, em))
            return TRUE;
        else
            return FALSE;
    case JDWP_EventRequest_Set_Out_modifiers_Modifier_ClassExclude:
        break;
    case JDWP_EventRequest_Set_Out_modifiers_Modifier_LocationOnly:
        if ((cep->loc.classID != em->u.loc.classID) ||
             (cep->loc.methodIndex != em->u.loc.methodIndex) ||
             (cep->loc.offset  != em->u.loc.offset))
             return FALSE;
        break;
    case JDWP_EventRequest_Set_Out_modifiers_Modifier_ExceptionOnly: {
        bool_t ret = TRUE;
        if (em->u.exp.classID != 0) {
            INSTANCE_CLASS clazz = (INSTANCE_CLASS)getObjectPtr(cep->exp.classID);
            while (clazz != NULL) {
                if (GET_CLASS_DEBUGGERID(&clazz->clazz) == em->u.exp.classID) {
                    ret = TRUE;
                    break;
                }
                clazz = clazz->superClass;
            }
        }

        ret = ret && (em->u.exp.sig_caught == cep->exp.sig_caught);
        ret = ret && (em->u.exp.sig_uncaught == cep->exp.sig_uncaught);

        if (ret == FALSE)
            return FALSE;

        break;
        }
    case JDWP_EventRequest_Set_Out_modifiers_Modifier_FieldOnly:
        break;
    case JDWP_EventRequest_Set_Out_modifiers_Modifier_Step:
        if ((cep->step.threadID != em->u.step.threadID) ||
            (cep->step.size != em->u.step.size) ||
            (cep->step.depth != em->u.step.depth)) {
    /*
    || (cep->step.target.cl != em->u.step.target.cl)
    || (cep->step.target.method != em->u.step.target.method)
    || (cep->step.target.offset != em->u.step.target.offset))
    */
             return FALSE;
        }
        break;
    }
    return TRUE;
}

/*========================================================================
 * Function:        checkEventCount()
 * Overview:        deals with events with a count modifier
 * Interface:
 * parameters:      event kind
 * returns:     nothing
 *
 * Notes:
 *=======================================================================*/

static void checkEventCount(VMEventPtr ep)
{
    if (ep->count_active) {
        if (ep->count == 0) {
            /* if it's 0 then the count was satisfied */
            ep->active = FALSE;
            ep->count_active = FALSE;
        }
    }
}

/*========================================================================
 * Function:        findSatisfyingEvent()
 * Overview:        finds a specific type of event that also satisfies
 *                  its modifiers
 * Interface:
 * parameters:      event kind, pointer to current event structure
 *            pointer to return list of events, count of events
 * returns:     true if event to process, false otherwise
 *
 * Notes:
 *=======================================================================*/

static bool_t findSatisfyingEvent(BYTE kind, CEModPtr cep, VMEventPtr *epp,
    int *eventCount, BYTE *suspendPolicy)
{
    VMEventPtr ep = eventHead;
    EVENTMODIFIER em;

    *epp = NULL;
    *eventCount = 0;
    *suspendPolicy = 0;
    if (ep == NIL)
        return (FALSE);     /* no events */
    do {
        if (ep->active == FALSE)
            continue;

        if (ep->kind == kind) {
            /*
             * The count modifier is a global modifier that filters all
             * other modifiers, hence it's in the VMEvent structure
             */
            if (ep->count_active) {
                if (ep->count-- > 1) {
                    continue;
                }
            }
        /* check modifiers */
        em = ep->mods;
        do  {
            if (em == NULL || satisfiesModifiers(ep, cep, em)) {
                /* add to list of events to send */
            ep->sendNext = NULL;
                if (*epp == NULL)  {
                    *epp = ep;
            } else {
                ep->sendNext = *epp;
                *epp = ep;
            }
            (*eventCount)++;
            if (ep->suspendPolicy > *suspendPolicy)
                *suspendPolicy = ep->suspendPolicy;
            break;
            }
            
        } while (em && (em = em->next) != NULL);
        }
    } while ((ep = ep->next));
    if (*epp != NULL)
        return TRUE;
    else
        return FALSE;
}

/*========================================================================
 * Function:        buildClassSignature()
 * Overview:        constructs a JNI signature from a class
 *                  
 * Interface:
 *  parameters:     class pointer, character buffer (previously alloc'd)
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

static void buildClassSignature(CLASS clazz, CONST_CHAR_HANDLE buf)
{
    char *str;

    if (IS_ARRAY_CLASS(clazz)) {
        str = getClassName(clazz);
        strcpy((char *)unhand(buf), str);
    } else {
        ((char *)(unhand(buf)))[ 0 ] = 'L'; /* typeCodeToSignature(T_CLASS); */
        ((char *)(unhand(buf)))[ 1 ] = 0x00;
        if (clazz->packageName != NIL) {
            strcat((char *)unhand(buf), UStringInfo(clazz->packageName));
            strcat((char *)unhand(buf), "/");
        }
        if (clazz->baseName != NIL)
            strcat((char *)unhand(buf), UStringInfo(clazz->baseName));

        strcat((char *)unhand(buf), ";");
    }
}

/*========================================================================
 * Function:        setNotification()
 * Overview:        tells the debugger to only listen for only 
 *                  one type of event
 *                  
 * Interface:
 *  parameters:     debugger event type
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

static void setNotification(long n)
{
    debuggerNotifyList = n;
}

/*========================================================================
 * Function:        addNotification()
 * Overview:        tells the debugger to listen to the event as well as
 *                  anything else that was set previously
 *                  
 * Interface:
 *  parameters:     debugger event type
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

static void addNotification(long n)
{
    debuggerNotifyList |= n;
}

/*========================================================================
 * Function:        FreeCEModifier()
 * Overview:        free a modifier structure for an event
 *                  
 * Interface:
 *  parameters:     
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void FreeCEModifier(CEModPtr cep) {

    cep->inUse = FALSE;
    cep->nextEvent = NULL;
}

/*========================================================================
 * Function:        GetCEModifier()
 * Overview:        get a free modifier structure for an event
 *                  
 * Interface:
 *  parameters:     
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

CEModPtr GetCEModifier() {

    CEModPtr cep = CE_ModHead;
    CEModPtr lastCep;

    while (cep) {
        if (!cep->inUse) {
            lastCep = cep->next;
            memset(cep, 0, sizeof(*cep));
            cep->next = lastCep;
            cep->inUse = TRUE;
            return (cep);
        }
        lastCep = cep;
        cep = cep->next;
    }
    cep = (CEModPtr)allocAlignedBytes(sizeof(*cep));
    if (CE_ModHead == NIL) {
            CE_ModHead = cep;
    } else {
        lastCep->next = cep;
    }
    cep->inUse = TRUE;
    return(cep);
}

/*========================================================================
 * Function:        getModPtr()
 * Overview:        get a free modifier structure for an event
 *                  
 * Interface:
 *  parameters:     
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

static EVENTMODIFIER getModPtr() {

    EVENTMODIFIER newMod = modHead;
    EVENTMODIFIER lastMod = NULL;

    while (newMod) {
        if (!newMod->inUse) {
            lastMod = newMod->freeNext;
            memset(newMod, 0, sizeof(*newMod));
            newMod->freeNext = lastMod;
            newMod->inUse = TRUE;
            return (newMod);
        }
        lastMod = newMod;
        newMod = newMod->freeNext;
    }
    newMod = (EVENTMODIFIER)allocAlignedBytes(sizeof(*newMod));
    if (modHead == NIL) {
            modHead = newMod;
    } else {
        lastMod->freeNext = newMod;
    }
    newMod->inUse = TRUE;
    return(newMod);
}

/*========================================================================
 * Function:        createVmEvent()
 * Overview:        creates an event structure but doesn't put it in
 *                  the list of events
 *                  
 * Interface:
 *  parameters:     
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

static VMEventPtr createVmEvent()
{
    VMEventPtr  ep = eventHead;
    VMEventPtr lastEp = NULL;

    /* first try to find an event to be reaped */
    while (ep != NULL) {
        if (!ep->inUse) {
            lastEp = ep->next;
            memset(ep, 0, sizeof(*ep));
            ep->next = lastEp;
            ep->active = TRUE;
            ep->mods = NIL;
            ep->numModifiers = 0;
            ep->inUse = TRUE;
            return (ep);
        }
        lastEp = ep;
        ep = ep->next;
    }
    ep = (VMEventPtr)callocPermanentObject(SIZEOF_VMEVENT);
    ep->active = TRUE;
    ep->mods = NIL;
    ep->numModifiers = 0;
    ep->inUse = TRUE;
    if (eventHead == NIL) {
            eventHead = ep;
    } else {
        lastEp->next = ep;
    }
    return ep;
}


/*========================================================================
 * Function:        installBreakpoint()
 * Overview:        set a breakpoint in the given method
 * Interface:
 *   parameters:    pointer to debug header, event type
 *   returns:       nothing
 *
 * Notes:       Packet format:
 *              Input:
 *                  suspend policy  byte
 *                  number of events    int
 *                  location        17 bytes
 *              Output:
 *                  debug header -      11 bytes
 *                  breakpoint ID   int
 *=======================================================================*/

static VMEventPtr installBreakpoint(PacketInputStream *in, PacketOutputStream *out, EVENTMODIFIER *mods)
{
    unsigned char bk = BREAKPOINT;
    unsigned long  classID;
    METHOD thisMethod;
    unsigned int offset;
    ByteCode opcode;
    char *start;
        INDICATE_DYNAMICALLY_INSIDE_TEMPORARY_ROOTS

    classID = (*mods)->u.loc.classID;
    IS_TEMPORARY_ROOT(thisMethod, getMethodFromIndex(classID, (*mods)->u.loc.methodIndex));
    offset = (*mods)->u.loc.offset;     

    if (thisMethod->u.java.code[offset] == BREAKPOINT) {
        findBreakEvent(&opcode, classID, (*mods)->u.loc.methodIndex,
           offset);
    }
    else
        opcode = thisMethod->u.java.code[offset];

#if ENABLEFASTBYTECODES
        /*
         * If this opcode is a fast opcode, we need to flush the
         * original byte code back into the code and then install
         * the breakpoint
         */
        if (opcode >= GETFIELD_FAST && opcode <= INSTANCEOF_FAST) {
            long iCacheIndex = getUShort(&(thisMethod->u.java.code[offset + 1]));
            releaseInlineCacheEntry(iCacheIndex);
            opcode = thisMethod->u.java.code[offset];
        }
#endif
    start = (char *)((INSTANCE_CLASS)getObjectPtr(classID))->constPool;
    setBreakpoint(thisMethod, offset, start, bk);
    (*mods)->u.loc.opcode = opcode;

    return 0;
}

/*========================================================================
 * Function:        processBreakCommands()
 * Overview:        called  after we send an event to debugger
 * Interface:
 *   parameters:    none
 *   returns:       nothing
 *
 * Notes:
 *=======================================================================*/

void processBreakCommands()
{
    resumeOK = FALSE;

    /* wait for some command to come from the debugger */
    for (;;) {
        ProcessDebugCmds(-1);   /* wait forever */
        if (resumeOK) {
            resumeOK = FALSE;
            return;
        }
    }
}

/*========================================================================
 * Function:        handleSuspendPolicy()
 * Overview:        determines which suspend policy this event was to
 *                  use and acts appropriately
 *                  
 * Interface:
 *  parameters:     policy, thread pointer, bool_t (unused)
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

static void handleSuspendPolicy(BYTE policy, THREAD thread, bool_t forceWait)
{
    switch(policy)
    {
        case JDWP_SuspendPolicy_NONE:
            /* do nothing */
            break;
        case JDWP_SuspendPolicy_EVENT_THREAD:
            suspendSpecificThread(thread);
            break;
        case JDWP_SuspendPolicy_ALL:
            waitOnSuspend = forceWait;
            VirtualMachine_Suspend(&inputStream, &outputStream);
            break;
    }
    
}

/*========================================================================
 * Function:        getMethodIndex()
 * Overview:        given a class and method address, find the index
 *          of that method in the methodtable
 *
 * Interface:
 *  parameters: Class, method
 *  returns:    int
 *
 * Notes:
 *=======================================================================*/

static long getMethodIndex(INSTANCE_CLASS cl, METHOD thisMethod) {

    int index = 0;
    FOR_EACH_METHOD(mp, cl->methodTable)
        if ((unsigned long)mp == (unsigned long)thisMethod)
            return index;
        index++;
    END_FOR_EACH_METHOD
    fatalVMError(KVM_MSG_DEBUGGER_COULD_NOT_FIND_METHOD);
    return 0;
}

/*========================================================================
 * Function:        getFieldAddress()
 * Overview:        given a class and field index, find the address
 *          of that method in the methodtable
 *
 * Interface:
 *  parameters: INSTANCE_CLASS, index
 *  returns:    FIELD
 *
 * Notes:
 *=======================================================================*/

static FIELD getFieldAddress(INSTANCE_CLASS cl, int index) {

    long offset = 0; 
    FIELDTABLE fieldTable = cl->fieldTable;
    /*  Get the length of the field table */ 
    FOR_EACH_FIELD(thisField, fieldTable) 
        if (offset == index) 
            return thisField;
        offset++;
    END_FOR_EACH_FIELD
           
    fatalVMError(KVM_MSG_DEBUGGER_COULD_NOT_FIND_FIELD);
    
    return NULL;
}

static void sendEvent(CEModPtr cep) {

    switch (cep->eventKind) {
    case JDWP_EventKind_CLASS_PREPARE:
        setEvent_ClassPrepare(cep);
        break;

    case JDWP_EventKind_CLASS_LOAD:
        setEvent_ClassLoad(cep);
        break;

    case JDWP_EventKind_THREAD_START:
        setEvent_ThreadStart(cep);
        break;
    }
}

void insertDebugEvent(CEModPtr cep) {

    CEModPtr lastcep;

    if (CurrentThread != NULL) {
        if (CurrentThread->debugEvent == NULL)
            CurrentThread->debugEvent = cep;
        else {
            lastcep = CurrentThread->debugEvent;
            while (lastcep->nextEvent != NULL) {
                lastcep = lastcep->nextEvent;
            }
            lastcep->nextEvent = cep;
        }
        CurrentThread->needEvent = TRUE;
    } else {
        sendEvent(cep);
        FreeCEModifier(cep);
    }
}

void checkDebugEvent(THREAD tptr) {

    CEModPtr cep;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(tptr, tptr);
    tptr->needEvent = FALSE;
    while ((cep = tptr->debugEvent) != NULL) {
        sendEvent(cep);
        tptr->debugEvent = cep->nextEvent;
        FreeCEModifier(cep);
    }
    END_TEMPORARY_ROOTS
    return;
}

/*========================================================================
 * Function:        setEvent_NeedSteppingInfo()
 * Overview:        tell the _proxy_ that _we_ need info about class lines
 *                  
 * Interface:
 *  parameters:         pointer to thread that's stepping
 *              pointer to CEMod structure
 *  returns:        void
 *
 * Notes:   Since this is called re-entrantly via createevent, we need
 *      to use our own output packet since on return to createevent
 *      and thence to processCommands, we don't want to have blown the
 *      global outputStream  used for most commands
 *=======================================================================*/

static void setEvent_NeedSteppingInfo(unsigned long threadID, CEModPtr cep)
{
    FRAME frame;
    unsigned long offset;
    THREAD thread;
    PacketOutputStream *out;
    unsigned long classID, methodIndex;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM, GCT_POINTERLIST));
    outStream_initCommand(out, uniqueID(), FLAGS_None, KVM_CMDSET,
        KVM_STEPPING_EVENT_COMMAND);

    IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(threadID));
    if (thread == CurrentThread) {
        frame =  fp_global;
        offset = ip_global - frame->thisMethod->u.java.code;
    } else {
        frame = thread->fpStore;
        offset = thread->ipStore - frame->thisMethod->u.java.code;
    }       
    classID = GET_CLASS_DEBUGGERID(&frame->thisMethod->ofClass->clazz);
    methodIndex = getMethodIndex(frame->thisMethod->ofClass, frame->thisMethod);
    outStream_writeLong(out, (long)cep);
    outStream_writeClassID(out, (CLASS)classID);
    outStream_writeLong(out, methodIndex);
    outStream_writeLong64(out, offset);
    outStream_sendCommand(out);
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        setEvent_VMInit()
 * Overview:        called when the vm starts
 *                  
 * Interface:
 *  parameters:     none
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void setEvent_VMInit()
{
    VMEventPtr reqID;
    PacketOutputStream *out;


    reqID = NIL;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM, GCT_POINTERLIST));
    outStream_initCommand(out, uniqueID(), FLAGS_None,
        JDWP_COMMAND_SET(Event), JDWP_COMMAND(Event, Composite));

    outStream_writeByte(out, JDWP_SuspendPolicy_ALL);
    outStream_writeLong(out, 1);
    outStream_writeByte(out, JDWP_EventKind_VM_START);
    outStream_writeLong(out, (long)reqID);
    /* hopefully this will be called when CurrentThread==initial_thread */
    outStream_writeObjectRef(out, (OBJECT)getObjectID(CURRENTTHREAD));
    outStream_sendCommand(out);
    
    handleSuspendPolicy(JDWP_SuspendPolicy_ALL, CurrentThread, TRUE);
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        setEvent_VMDeath()
 * Overview:        called when the vm terminates
 *                  
 * Interface:
 *  parameters:     none
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void setEvent_VMDeath()
{
    PacketOutputStream *out;

    checkNOTIFY_WANTED(Dbg_EventKind_VM_DEATH);

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM, GCT_POINTERLIST));
    outStream_initCommand(out, uniqueID(), FLAGS_None,
        JDWP_COMMAND_SET(Event), JDWP_COMMAND(Event, Composite));

    outStream_writeByte(out, JDWP_SuspendPolicy_ALL);
    outStream_writeLong(out, 1);
    outStream_writeByte(out, JDWP_EventKind_VM_DEATH);
    outStream_writeLong(out, 0); /* spec says always zero */
    outStream_sendCommand(out);
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        setEvent_SingleStep
 * Overview:        
 *                  
 * Interface:
 *  parameters:     none
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void setEvent_SingleStep(CEModPtr cep)
{
    VMEventPtr ep, epFirst;
    int eventCount = 0;
    BYTE suspendPolicy = JDWP_SuspendPolicy_NONE;
    PacketOutputStream *out;

    checkNOTIFY_WANTED(Dbg_EventKind_SINGLE_STEP);
        /* In case of an allocation in outStream_writeBytes */
    if (findSatisfyingEvent(JDWP_EventKind_SINGLE_STEP, cep, &epFirst,
        &eventCount, &suspendPolicy)) {
        ep = epFirst;
        START_TEMPORARY_ROOTS
        IS_TEMPORARY_ROOT(out, (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM, GCT_POINTERLIST));
        outStream_initCommand(out, uniqueID(), FLAGS_None,
        JDWP_COMMAND_SET(Event), JDWP_COMMAND(Event, Composite));
        outStream_writeByte(out, suspendPolicy);
        outStream_writeLong(out, eventCount);
        while (ep != NULL) {
            outStream_writeByte(out, JDWP_EventKind_SINGLE_STEP);
            outStream_writeObjectRef(out, (OBJECT)ep);
            outStream_writeObjectRef(out, (OBJECT)getObjectID(CURRENTTHREAD));
            outStream_writeLocation(out, JDWP_TypeTag_CLASS,
                (CLASS)cep->loc.classID, (METHOD)cep->loc.methodIndex,
            cep->loc.offset);

        checkEventCount(ep);
        ep = ep->sendNext;
        }
            outStream_sendCommand(out);
        handleSuspendPolicy(suspendPolicy, CurrentThread, FALSE);
        END_TEMPORARY_ROOTS
    }
}

/*========================================================================
 * Function:        setEvent_Breakpoint()
 * Overview:        sends a break event to debugger
 * Interface:
 *   parameters:    pointer to breakpoint structure, type of event
 *   returns:       nothing
 *
 * Notes:       writes out a composite event packet per JDWP spec
 *=======================================================================*/

void setEvent_Breakpoint(VMEventPtr ep, CEModPtr cep)
{
    EVENTMODIFIER em;
    PacketOutputStream *out;

    /* BYTE suspendPolicy = JDWP_SuspendPolicy_ALL; */

    if (!NOTIFY_WANTED(Dbg_EventKind_SINGLE_STEP) &&
         !NOTIFY_WANTED(Dbg_EventKind_BREAKPOINT))
         return;

    em = ep->mods;
    if (satisfiesModifiers(ep, cep, em) == FALSE)
        return; 

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM, GCT_POINTERLIST));
    outStream_initCommand(out, uniqueID(), FLAGS_None,
        JDWP_COMMAND_SET(Event), JDWP_COMMAND(Event, Composite));
    outStream_writeByte(out, ep->suspendPolicy);

    /* count is 1 for now */
    outStream_writeLong(out, 1);
    outStream_writeByte(out, JDWP_EventKind_BREAKPOINT);
    outStream_writeLong(out, (long)ep);
    outStream_writeObjectRef(out, (OBJECT)getObjectID(CURRENTTHREAD));

    /* write the Location */
    outStream_writeLocation(out, JDWP_TypeTag_CLASS,
        (CLASS)cep->loc.classID, (METHOD)cep->loc.methodIndex,
        cep->loc.offset);

    outStream_sendCommand(out);
    handleSuspendPolicy(ep->suspendPolicy, CurrentThread, FALSE);
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        setEvent_MethodEntry()
 * Overview:        called when the code enters a method
 *                  
 * Interface:
 *  parameters:     none
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void setEvent_MethodEntry()
{
    checkNOTIFY_WANTED(Dbg_EventKind_METHOD_ENTRY);
}

/*========================================================================
 * Function:        setEvent_MethodExit()
 * Overview:        called when the code exits a method
 *                  
 * Interface:
 *  parameters:     
 *  returns:        
 *
 * Notes:
 *=======================================================================*/

void setEvent_MethodExit()
{
    checkNOTIFY_WANTED(Dbg_EventKind_METHOD_EXIT);
}

/*========================================================================
 * Function:        setEvent_ThreadStart()
 * Overview:        called when a thread is created
 *                  
 * Interface:
 *  parameters:     none
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void setEvent_ThreadStart(CEModPtr cep)
{
    VMEventPtr ep, epFirst;
    int eventCount = 0;
    BYTE suspendPolicy = JDWP_SuspendPolicy_NONE;
    PacketOutputStream *out;


    checkNOTIFY_WANTED(Dbg_EventKind_THREAD_START);

    if (findSatisfyingEvent(JDWP_EventKind_THREAD_START, cep, &epFirst,
    &eventCount, &suspendPolicy)) {
        ep = epFirst;
        START_TEMPORARY_ROOTS
        IS_TEMPORARY_ROOT(out, (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM, GCT_POINTERLIST));
        outStream_initCommand(out, uniqueID(), FLAGS_None,
            JDWP_COMMAND_SET(Event), JDWP_COMMAND(Event, Composite));
        outStream_writeByte(out, suspendPolicy);
        outStream_writeLong(out, eventCount);
        while(ep != NULL) {
            outStream_writeByte(out, JDWP_EventKind_THREAD_START);
            outStream_writeLong(out, (long)ep);
           outStream_writeObjectRef(out, (OBJECT)cep->threadID);
        
        checkEventCount(ep);
        ep = ep->sendNext;
        }
        outStream_sendCommand(out);
        handleSuspendPolicy(suspendPolicy, GET_VMTHREAD(cep->threadID), FALSE);
        END_TEMPORARY_ROOTS
    }
}

/*========================================================================
 * Function:        setEvent_ThreadDeath()
 * Overview:        called when a thread terminates
 *                  
 * Interface:
 *  parameters:     
 *  returns:        
 *
 * Notes:
 *=======================================================================*/

void setEvent_ThreadDeath(CEModPtr cep)
{
    VMEventPtr ep, epFirst;
    int eventCount = 0;
    BYTE suspendPolicy = JDWP_SuspendPolicy_NONE;
    PacketOutputStream *out;


    checkNOTIFY_WANTED(Dbg_EventKind_THREAD_END);

    if (findSatisfyingEvent(JDWP_EventKind_THREAD_END, cep, &epFirst,
    &eventCount, &suspendPolicy)) {    
    ep = epFirst;
        START_TEMPORARY_ROOTS
        IS_TEMPORARY_ROOT(out, (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM, GCT_POINTERLIST));
        outStream_initCommand(out, uniqueID(), FLAGS_None,
            JDWP_COMMAND_SET(Event), JDWP_COMMAND(Event, Composite));
        outStream_writeByte(out, suspendPolicy);
        outStream_writeLong(out, eventCount);
    while (ep != NULL) {
            outStream_writeByte(out, JDWP_EventKind_THREAD_END);
            outStream_writeLong(out, (long)ep);
         outStream_writeObjectRef(out, (OBJECT)cep->threadID);
        checkEventCount(ep);
        ep = ep->sendNext;
        }
        outStream_sendCommand(out);
        handleSuspendPolicy(suspendPolicy, GET_VMTHREAD(cep->threadID), FALSE);
        END_TEMPORARY_ROOTS
    }
}

/*========================================================================
 * Function:        getJDWPClassStatus()
 * Overview:        translates an internal class status 
 *                  into the JDWP equivalent.
 *                  
 * Interface:
 *  parameters:     class pointer
 *  returns:        JDWP class status
 *
 * Notes:
 *=======================================================================*/

int getJDWPClassStatus(INSTANCE_CLASS clazz)
{
    int ret = 0;

    if (IS_ARRAY_CLASS(clazz))
        return JDWP_ClassStatus_INITIALIZED;
    switch (clazz->status) {

    case CLASS_ERROR:
        return JDWP_ClassStatus_ERROR;
    /* case CLASS_RAW: */
    case CLASS_READY:
        ret = JDWP_ClassStatus_INITIALIZED;
    /* fall through */
    case CLASS_LOADING:
    case CLASS_LOADED:
    case CLASS_LINKED:
        ret |= JDWP_ClassStatus_PREPARED;
    /* fall through */
    case CLASS_VERIFIED:
        ret |= JDWP_ClassStatus_VERIFIED;
        break;
    }
    return ret;
}

/*========================================================================
 * Function:        getJDWPThreadStatus()
 * Overview:        translates an internal thread status
 *                  into the JDWP equivalent
 *                  
 * Interface:
 *  parameters:     thread pointer
 *  returns:        JDWP thread status
 *
 * Notes:
 *=======================================================================*/

static int getJDWPThreadStatus(THREAD thread)
{
    int state = thread->state;
    if (state == THREAD_SUSPENDED) {
        return JDWP_ThreadStatus_RUNNING;
    } 
    else {
        state &= ~THREAD_SUSPENDED;
        state &= ~THREAD_DBG_SUSPENDED;
        switch (state) {
            case THREAD_ACTIVE:
                return JDWP_ThreadStatus_RUNNING;
            case THREAD_MONITOR_WAIT:
                return JDWP_ThreadStatus_MONITOR;
            case THREAD_CONVAR_WAIT:
                return JDWP_ThreadStatus_WAIT;
            case THREAD_DEAD:
                return JDWP_ThreadStatus_ZOMBIE;
            default:
                return JDWP_ThreadStatus_UNKNOWN;
        }
    }
}

/*========================================================================
 * Function:        getJDWPThreadSuspendStatus()
 * Overview:        returns if this thread has been suspended 
 *                  through the debugger
 *                  
 * Interface:
 *  parameters:     thread pointer
 *  returns:        suspend status
 *
 * Notes:
 *=======================================================================*/

static int getJDWPThreadSuspendStatus(THREAD thread) 
{
    int ret = 0;
    int state = thread->state;
    
    if (state == THREAD_SUSPENDED) 
        ret |= JDWP_SuspendStatus_SUSPEND_STATUS_SUSPENDED;

    if (state & THREAD_DBG_SUSPENDED) 
        ret |= JDWP_SuspendStatus_SUSPEND_STATUS_SUSPENDED;
        
    if ((state & THREAD_MONITOR_WAIT) || (state & THREAD_CONVAR_WAIT))
        ret |= JDWP_SuspendStatus_SUSPEND_STATUS_SUSPENDED;

    return ret;
}

/*========================================================================
 * Function:        getJDWPTagType()
 * Overview:        determines what kind of class a class is
 *                  (array, interface, class)
 *                  
 * Interface:
 *  parameters:     class pointer
 *  returns:        tag type
 *
 * Notes:
 *=======================================================================*/

static BYTE getJDWPTagType(CLASS clazz) 
{
    if (IS_ARRAY_CLASS(clazz))
        return JDWP_TypeTag_ARRAY;
    else if (clazz->accessFlags & ACC_INTERFACE)
        return JDWP_TypeTag_INTERFACE;
    else /* is a class */
        return JDWP_TypeTag_CLASS;
}

/*========================================================================
 * Function:        sendClassInfo
 * Overview:        sends information about a class to the debugger
 *                  (i.e. tag type, class id, status, and (maybe) sig)
 *                  
 * Interface:
 *  parameters:     output pointer, class pointer, include sig bool_t
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

static void sendClassInfo(PacketOutputStream *out, CLASS clazz,
    CONST_CHAR_HANDLE buffer, bool_t incSig)
{
    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    outStream_writeByte(out, getJDWPTagType(clazz));
    outStream_writeClassID(out, (CLASS)(GET_CLASS_DEBUGGERID(clazz)));

    if (incSig == TRUE) {
        buildClassSignature(clazz, buffer);
        outStream_writeString(out, buffer);
    }

    if (IS_ARRAY_CLASS(clazz))
        outStream_writeLong(out, JDWP_ClassStatus_INITIALIZED);
    else
        outStream_writeLong(out, getJDWPClassStatus((INSTANCE_CLASS)clazz));
    END_TEMPORARY_ROOTS
}

void sendAllClassPrepares() {

    CEModPtr cep;
    checkNOTIFY_WANTED(Dbg_EventKind_CLASS_PREPARE);
    cep = GetCEModifier();
    FOR_ALL_CLASSES(clazz)
        cep->loc.classID = GET_CLASS_DEBUGGERID(clazz);
        cep->threadID = getObjectID(CURRENTTHREAD);
        setEvent_ClassPrepare(cep);
    END_FOR_ALL_CLASSES
    FreeCEModifier(cep);
}

/*========================================================================
 * Function:        setEvent_ClassPrepare()
 * Overview:        called when a class is loading
 *                  
 * Interface:
 *  parameters:     thread pointer, class pointer
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void setEvent_ClassPrepare(CEModPtr cep)
{
    CLASS clazz = getObjectPtr(cep->loc.classID);
    VMEventPtr ep, epFirst;
    int eventCount = 0;
    char *buffer;
    BYTE suspendPolicy = JDWP_SuspendPolicy_NONE;
    PacketOutputStream *out;

    checkNOTIFY_WANTED(Dbg_EventKind_CLASS_PREPARE);

    if (!IS_ARRAY_CLASS(clazz) &&
        ((INSTANCE_CLASS)clazz)->status < CLASS_LOADED)
            return;
        /* In case of an allocation in outStream_writeBytes */
        START_TEMPORARY_ROOTS
        if (findSatisfyingEvent(JDWP_EventKind_CLASS_PREPARE, cep, &epFirst,
        &eventCount, &suspendPolicy)) {
            ep = epFirst;
            IS_TEMPORARY_ROOT(out, (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM, GCT_POINTERLIST));
            outStream_initCommand(out, uniqueID(), FLAGS_None,
                JDWP_COMMAND_SET(Event), JDWP_COMMAND(Event, Composite));
            outStream_writeByte(out, suspendPolicy);
            outStream_writeLong(out, eventCount);
            IS_TEMPORARY_ROOT(buffer, mallocBytes(256));
            while (ep != NULL) {
                outStream_writeByte(out, JDWP_EventKind_CLASS_PREPARE);        
                outStream_writeLong(out, (long)ep);
                outStream_writeObjectRef(out, (OBJECT)cep->threadID);

                outStream_writeByte(out, getJDWPTagType(clazz));
                outStream_writeClassID(out, (CLASS)(GET_CLASS_DEBUGGERID(clazz)));

                buildClassSignature(clazz, (CONST_CHAR_HANDLE)&buffer);

                outStream_writeString(out, (CONST_CHAR_HANDLE)&buffer);
                outStream_writeLong(out,
                    getJDWPClassStatus((INSTANCE_CLASS)clazz));

                checkEventCount(ep);
                ep = ep->sendNext;
            }
            outStream_sendCommand(out);
            handleSuspendPolicy(suspendPolicy, GET_VMTHREAD(cep->threadID), FALSE);
        }
        END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        setEvent_ClassLoad()
 * Overview:        called when a class is loaded
 *                  
 * Interface:
 *  parameters:     thread pointer, class pointer
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void setEvent_ClassLoad(CEModPtr cep)
{
    checkNOTIFY_WANTED(Dbg_EventKind_CLASS_LOAD);
    setEvent_ClassPrepare(cep);
}

/*========================================================================
 * Function:        setEvent_ClassUnload()
 * Overview:        called when a class is unloaded
 *                  
 * Interface:
 *  parameters:     thread pointer, class pointer
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void setEvent_ClassUnload(CEModPtr cep)
{
    CLASS clazz = getObjectPtr(cep->loc.classID);
    VMEventPtr ep, epFirst;
    int eventCount = 0;
    BYTE suspendPolicy = JDWP_SuspendPolicy_NONE;
    char *buffer;
    PacketOutputStream *out;


    checkNOTIFY_WANTED(Dbg_EventKind_CLASS_UNLOAD);

        /* In case of an allocation in outStream_writeBytes */
        START_TEMPORARY_ROOTS
        if (findSatisfyingEvent(JDWP_EventKind_CLASS_UNLOAD, cep, &epFirst,
        &eventCount, &suspendPolicy)) {
        ep = epFirst;
        IS_TEMPORARY_ROOT(out, (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM, GCT_POINTERLIST));
            outStream_initCommand(out, uniqueID(), FLAGS_None,
                JDWP_COMMAND_SET(Event), JDWP_COMMAND(Event, Composite));
            outStream_writeByte(out, suspendPolicy);
            outStream_writeLong(out, eventCount);
        IS_TEMPORARY_ROOT(buffer, mallocBytes(256));
            while (ep != NULL) {
                outStream_writeByte(out, JDWP_EventKind_CLASS_UNLOAD);
                outStream_writeLong(out, (long)ep);
                buildClassSignature(clazz, (CONST_CHAR_HANDLE)&buffer);

                outStream_writeString(out, (CONST_CHAR_HANDLE)&buffer);
                checkEventCount(ep);
        ep = ep->sendNext;
            }
            outStream_sendCommand(out);
            handleSuspendPolicy(suspendPolicy, GET_VMTHREAD(cep->threadID), FALSE);
        }
        END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        setEvent_FieldAccess()
 * Overview:        called when a field is accessed (read)
 *                  
 * Interface:
 *  parameters:     none
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void setEvent_FieldAccess()
{
    checkNOTIFY_WANTED(Dbg_EventKind_FIELD_ACCESS);
}

/*========================================================================
 * Function:        setEvent_FieldModification()
 * Overview:        called when a field is modified
 *                  
 * Interface:
 *  parameters:     
 *  returns:        
 *
 * Notes:
 *=======================================================================*/

void setEvent_FieldModification()
{
    checkNOTIFY_WANTED(Dbg_EventKind_FIELD_MODIFICATION);
}

/*========================================================================
 * Function:        setEvent_FramePop()
 * Overview:        
 *                  
 * Interface:
 *  parameters:     
 *  returns:        
 *
 * Notes:
 *=======================================================================*/

void setEvent_FramePop()
{
    checkNOTIFY_WANTED(Dbg_EventKind_FRAME_POP);
}

/*========================================================================
 * Function:        setEvent_Exception()
 * Overview:        
 *                  
 * Interface:
 *  parameters:     
 *  returns:        
 *
 * Notes:
 *=======================================================================*/

void setEvent_Exception(INSTANCE exception, FRAME throwFrame, BYTE *throwIP,
    METHOD catchMethod, unsigned long catchOffs, CEModPtr cep)
{
    VMEventPtr ep, epFirst;
    THREAD thread = GET_VMTHREAD(cep->threadID);
    int eventCount = 0;
    BYTE suspendPolicy = JDWP_SuspendPolicy_NONE;
    PacketOutputStream *out;

    
    checkNOTIFY_WANTED(Dbg_EventKind_EXCEPTION);
    
    if (thread->javaThread == NIL) /* this isn't a java exception */
        return;
        /* In case of an allocation in outStream_writeBytes */
        START_TEMPORARY_ROOTS
    if (findSatisfyingEvent(JDWP_EventKind_EXCEPTION, cep, &epFirst,
        &eventCount, &suspendPolicy)) {
        IS_TEMPORARY_ROOT(exception, exception);
        IS_TEMPORARY_ROOT(throwFrame, throwFrame);
        IS_TEMPORARY_ROOT(catchMethod, catchMethod);
        IS_TEMPORARY_ROOT(throwIP, throwIP);
        ep = epFirst;
        IS_TEMPORARY_ROOT(out, (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM, GCT_POINTERLIST));
            outStream_initCommand(out, uniqueID(), FLAGS_None,
                JDWP_COMMAND_SET(Event), JDWP_COMMAND(Event, Composite));
            outStream_writeByte(out, suspendPolicy);
            outStream_writeLong(out, eventCount);
            while (ep != NULL) {
                outStream_writeByte(out, JDWP_EventKind_EXCEPTION);        
            
                outStream_writeLong(out, (long)ep);
            
                /* thread with exception */
                outStream_writeObjectRef(out, (OBJECT)cep->threadID);
            
                /* location of exception throw */
                outStream_writeLocation(out, JDWP_TypeTag_CLASS, 
                    (CLASS)(GET_CLASS_DEBUGGERID(&throwFrame->thisMethod->ofClass->clazz)),
                    (METHOD)getMethodIndex( 
                    throwFrame->thisMethod->ofClass,
                (METHOD)throwFrame->thisMethod),
                    (unsigned long)(throwIP - throwFrame->thisMethod->u.java.code));
    
                /* thrown exception */
                outStream_writeByte(out, 'L');
                outStream_writeObjectRef(out, (OBJECT)getObjectID(exception));
            
                /* location of catch, or 0 if not caught */
                if (catchMethod == NIL) {
                    outStream_writeLocation(out, 0, 0, 0, 0);
                } else {
                    outStream_writeLocation(out, JDWP_TypeTag_CLASS,
                        (CLASS)(GET_CLASS_DEBUGGERID(&catchMethod->ofClass->clazz)),
                        (METHOD)getMethodIndex(catchMethod->ofClass,
                        catchMethod), catchOffs);
                }
            
                checkEventCount(ep);
        ep = ep->sendNext;
            }
            outStream_sendCommand(out);
            handleSuspendPolicy(suspendPolicy, GET_VMTHREAD(cep->threadID), FALSE);
    }
        END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        setEvent_ExceptionCatch()
 * Overview:        
 *                  
 * Interface:
 *  parameters:     
 *  returns:        
 *
 * Notes:
 *=======================================================================*/

void setEvent_ExceptionCatch()
{
    checkNOTIFY_WANTED(Dbg_EventKind_EXCEPTION_CATCH);
}

/*========================================================================
 * Function:        getVerifierBreakpointOpcode()
 * Overview:        called by verifier loop to retrieve the underlying
 *          opcode that was breakpointed.
 * Interface:
 *   parameters:    none
 *   returns:       byte opcode
 *
 * Notes:
 *=======================================================================*/

ByteCode getVerifierBreakpointOpcode(METHOD thisMethod, unsigned short insp)
{
    INSTANCE_CLASS cl;
    ByteCode opcode;

    cl = thisMethod->ofClass;   /* point to class of method */
    if (findBreakEvent(&opcode, GET_CLASS_DEBUGGERID(&cl->clazz),
        getMethodIndex(cl, thisMethod), insp) == NIL) {
        raiseException("java/lang/VirtualMachineError");    /* XXX */
        return 0;
    }
    /*opcode = mod->u.loc.opcode;*/
    return opcode;
}

/*========================================================================
 * Function:        handleSingleStep()
 * Overview:        determines when a stepping event has occured
 *                  stepping can be in, over, or out.
 *                  
 * Interface:
 *  parameters:     
 *  returns:        
 *
 * Notes:
 *=======================================================================*/

int handleSingleStep(THREAD thread, THREAD *tp)
{
    unsigned long offset;
    FRAME frame, tframe;
    unsigned long classID;
    unsigned long methodIndex;
    bool_t sendEvent = FALSE;
    
    /* if the proxy has returned -1 it means that we have reached
     * the end of a function and we must look up stack for the
     * calling function
     *
     */
    frame = fp_global;
    offset = ip_global - frame->thisMethod->u.java.code;
    if (thread->stepInfo.target.offset == -1) {
    /*
     * Wait for the frame to change and that will mean we popped up to
     * the calling frame. (Or we called down to another frame, maybe...
     * should look into that possibility)
     */
    if (frame != thread->stepInfo.fp)
            sendEvent = TRUE;
    } else {

        switch (thread->stepInfo.size) {

        case JDWP_StepSize_LINE:

            /* stepping by line can have two effects: we stay in the same 
             * frame or the frame changes
             */

            if (thread->stepInfo.fp == frame) {

                /* if the frame is the same and we are not running to the
                 * end of a function we need to see if we have reached
                 * the location we are looking for
                 */
                if (thread->stepInfo.depth != JDWP_StepDepth_OUT) {

                    /* we reached the exact offset we were looking for */
                    if (offset == thread->stepInfo.target.offset) {
                        sendEvent = TRUE;
                    /* we ended up before the offset were the stepping
                     * request was initiated. we jumped back
                     */
                    } else if (offset < thread->stepInfo.offset) {
                        sendEvent = TRUE;
                    /* we are after the location we were looking for;
                     * we jumped forward
                     */
                    } else if (offset > thread->stepInfo.target.offset && 
                                offset >= thread->stepInfo.nextLineOffs) {
                        /* this may need more work */
                        sendEvent = TRUE;
                    }
                }

            } else {

                /* the frame has changed but we don't know whether we
                 * have gone into a function or gone up one.
                 * if we look up the call stack and find the frame 
                 * that we originally set the request then we stepped
                 * into a function, otherwise we stepped out 
                 */
                sendEvent = TRUE;
                if (thread->stepInfo.depth == JDWP_StepDepth_OVER ||
                     thread->stepInfo.depth == JDWP_StepDepth_OUT) {
                    for (tframe = frame->previousFp; tframe != NIL;
                tframe = tframe->previousFp) {
                        if (tframe == thread->stepInfo.fp) {
                             sendEvent = FALSE;
                             break;
                        }
                    }
                }
            }

            break;

        /* stepping by bytecode */
        case JDWP_StepSize_MIN:
            sendEvent = TRUE;
            break;
        }
    }

    if (sendEvent == TRUE) {
        CEModPtr cep;
        START_TEMPORARY_ROOTS
        IS_TEMPORARY_ROOT(thread, thread);
        classID = GET_CLASS_DEBUGGERID(&frame->thisMethod->ofClass->clazz);
        methodIndex = getMethodIndex(frame->thisMethod->ofClass, frame->thisMethod);
        cep = GetCEModifier();
        cep->step.threadID = getObjectID(thread->javaThread);
        cep->loc.classID = classID;
        cep->loc.methodIndex = methodIndex;
        cep->loc.offset = offset;
        cep->step.depth = thread->stepInfo.depth;
        cep->step.size = thread->stepInfo.size;

        setEvent_SingleStep(cep);
        FreeCEModifier(cep);
        *tp = thread;
        END_TEMPORARY_ROOTS
        if (CurrentThread == NIL) {
            signalTimeToReschedule();
            /* make sure that we don't continue executing
             * code in the CurrentThread if we have 
             * suspended it
             */
        return 1;
        }
    } else
        *tp = thread;
    return 0;
}

void handleBreakpoint(THREAD thread)
{
    VMEventPtr ep;
    INSTANCE_CLASS clazz;
    unsigned long offset;
    ByteCode opcode;
    CEModPtr cep;

    START_TEMPORARY_ROOTS
    DECLARE_TEMPORARY_ROOT(METHOD, thisMethod, fp_global->thisMethod);
    IS_TEMPORARY_ROOT(thread, thread);
    IS_TEMPORARY_ROOT(clazz, thisMethod->ofClass);
    offset = ip_global - thisMethod->u.java.code;   /* where are we? */
    
    /* find this breakpoint */
    ep = findBreakEvent(&opcode, GET_CLASS_DEBUGGERID(&clazz->clazz),
        getMethodIndex(clazz, thisMethod), offset);
    
    if (ep == NIL) {
        raiseException("java/lang/VirtualMachineError"); /* XXX */
        goto breakout;
    }

    thread->nextOpcode = opcode;
    thread->isAtBreakpoint = TRUE;

    cep = GetCEModifier();
    cep->loc.classID = GET_CLASS_DEBUGGERID(&clazz->clazz);
    cep->threadID = getObjectID(thread->javaThread);
    cep->loc.methodIndex = getMethodIndex(clazz, thisMethod);
    cep->loc.offset = offset;

    setEvent_Breakpoint(ep, cep);
    FreeCEModifier(cep);
        /*
         * Make sure that we don't continue executing
         * code in the CurrentThread if we have suspended it
         */
    signalTimeToReschedule();
breakout:
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        VirtualMachine_AllClasses()
 * Overview:        send all loaded class id's to the debugger
 * Interface:
 *   parameters:    pointer to debug header
 *   returns:       none
 *                  
 *
 * Notes:
 *=======================================================================*/

void VirtualMachine_AllClasses(PacketInputStream *in, PacketOutputStream *out)
{
    long cnt = 0;
    char *buffer;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    FOR_ALL_CLASSES(clazz)
        if (IS_ARRAY_CLASS(clazz) ||
            ((INSTANCE_CLASS)clazz)->status >= CLASS_LOADED) {
            if (inCurrentHeap(clazz) && GET_CLASS_DEBUGGERID(clazz) == 0) {
#if INCLUDEDEBUGCODE
                fatalError("Debugger: Class object ID is 0");
#endif
            }
            cnt++;
        }
    END_FOR_ALL_CLASSES

    outStream_writeLong(out, cnt);

    IS_TEMPORARY_ROOT(buffer, mallocBytes(256));
    FOR_ALL_CLASSES(clazz)
        if (IS_ARRAY_CLASS(clazz) ||
            ((INSTANCE_CLASS)clazz)->status >= CLASS_LOADED)
            sendClassInfo(out, clazz, (CONST_CHAR_HANDLE)&buffer, TRUE);
    END_FOR_ALL_CLASSES

    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        VirtualMachine_AllThreads()
 * Overview:        send all running thread id's to the debugger
 * Interface:
 *   parameters:    pointer to debug header
 *   returns:       none
 *                  
 *
 * Notes:
 *=======================================================================*/

void VirtualMachine_AllThreads(PacketInputStream *in, PacketOutputStream *out)
{
    THREAD tptr;
    
    long count=0;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    for (tptr=AllThreads; tptr; tptr = tptr->nextAliveThread) {
        if (!(tptr->state & THREAD_DEAD) &&
               !(tptr->state & THREAD_JUST_BORN)) {
            count++;
        }
    }

    outStream_writeLong(out, count);
    IS_TEMPORARY_ROOT(tptr, AllThreads);
    for (; tptr; tptr = tptr->nextAliveThread) {
        if (!(tptr->state & THREAD_DEAD) &&
               !(tptr->state & THREAD_JUST_BORN)) {
            outStream_writeObjectRef(out, (OBJECT)getObjectID(tptr->javaThread));
        }
    }
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        VirtualMachine_Suspend()
 * Overview:        called from interpreter loop to suspend VM
 * Interface:
 *   parameters:    none
 *   returns:       none
 *
 * Notes:
 *=======================================================================*/

void VirtualMachine_Suspend(PacketInputStream *in, PacketOutputStream *out)
{
    THREAD tptr;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(tptr, AllThreads);
    for (; tptr != NIL; tptr = tptr->nextAliveThread) {
        suspendSpecificThread(tptr);
    }
    allThreadsSuspended = TRUE;
    if (waitOnSuspend)
        processBreakCommands();
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        VirtualMachine_Resume()
 * Overview:        resume VM
 * Interface:
 *   parameters:    none
 *   returns:       none
 *
 * Notes:
 *=======================================================================*/

void VirtualMachine_Resume(PacketInputStream *in, PacketOutputStream *out)
{
    THREAD tptr;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    CurrentThread = NIL;
    IS_TEMPORARY_ROOT(tptr, AllThreads);
    for (; tptr != NIL; tptr = tptr->nextAliveThread) {
        resumeSpecificThread(tptr);
    }
    allThreadsSuspended = FALSE;
    setResumeOK(TRUE);
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        VirtualMachine_Exit()
 * Overview:        causes the vm to terminate
 *                  
 * Interface:
 *  parameters:     input pointer, output pointer, isBreak bool
 *  returns:        
 *
 * Notes:
 *=======================================================================*/

void VirtualMachine_Exit(PacketInputStream *in, PacketOutputStream *out)
{
    long errorCode;
    
    errorCode = inStream_readLong(in);
    
    setNotification(Dbg_EventKind_NONE);
    
    
    /* 
     * Stop the VM 
     */
    
    ERROR_THROW(errorCode);

}

static void getStaticValue(PacketOutputStream *out, FIELD fld, char typeTag) {


    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    if (inCurrentHeap(fld))
        IS_TEMPORARY_ROOT(fld, fld);
    if (fld->accessFlags & ACC_DOUBLE) {
            outStream_writeValue(out, typeTag,
            (void *)(fld->u.staticAddress),
            (void *)(&(((cell *)(fld->u.staticAddress))[1])), 1);
    } else {
            if (isObjectTag(typeTag))
                outStream_writeObjectValue(out, (OBJECT)(*(OBJECT *)fld->u.staticAddress), typeTag);
            else
                outStream_writeValue(out, typeTag,
                    (void *)(fld->u.staticAddress),
                    (void *)0, 1);
    }
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        ReferenceType_GetValues()
 * Overview:        get the value of one of an static variable
 * Interface:
 *   parameters:    pointer to debug header
 *   returns:
 *
 * Notes:       Packet format:
 *              Input:
 *                  class ID    Long
 *                  number of slots     Long
 *    
 *              Output:
 *                  debug header -      11 bytes
 *                  count - Long
 *                  taggedvalues - byte + Long
 *=======================================================================*/
 
void ReferenceType_GetValues(PacketInputStream *in, PacketOutputStream *out)
{
    INSTANCE_CLASS clazz, fieldClazz;
    unsigned long classID, fieldClassID;
    long num;
    FIELD fld = NULL;
    int i;
    char typeTag;
    char *sig;
    long fieldID;
    long zero = 0;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    IS_TEMPORARY_ROOT(sig, mallocBytes(DEBUG_MAX_SIG));
    classID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(clazz,  (INSTANCE_CLASS)getObjectPtr(classID));
    num = inStream_readLong(in);
    outStream_writeLong(out, num);
    for (i = 0; i < num; i++) {
        fieldClassID = inStream_readLong(in);
        fieldID = inStream_readLong(in);
        IS_TEMPORARY_ROOT(fieldClazz, (INSTANCE_CLASS)getObjectPtr(fieldClassID));
        if (fieldClazz == NULL || fieldClazz->fieldTable == NULL) {
            outStream_setError(out, JDWP_Error_INVALID_FIELDID);
            goto breakout;
        }
        if (fieldID >= fieldClazz->fieldTable->length) {
            outStream_setError(out, JDWP_Error_INVALID_FIELDID);
            goto breakout;
        }
        fld = getFieldAddress(fieldClazz, fieldID);
        START_TEMPORARY_ROOTS
        if (inCurrentHeap(fld))
            IS_TEMPORARY_ROOT(fld, fld);
        change_Key_to_FieldSignature_inBuffer(
            fld->nameTypeKey.nt.typeKey, sig);
        typeTag = *sig;
        if (!(fld->accessFlags & ACC_STATIC)) {
            if (fld->accessFlags & ACC_DOUBLE) {
                outStream_writeValue(out, typeTag,
                (void *)(&zero), (void *)(&zero), 1);
            } else {
                if (isObjectTag(typeTag))
                    outStream_writeObjectValue(out, NULL, typeTag);
                else
                    outStream_writeValue(out, typeTag,
                        (void *)(&zero), (void *)0, 1);
           }
        } else {
            getStaticValue(out, fld, typeTag);
        }
        END_TEMPORARY_ROOTS
    }
breakout:
    END_TEMPORARY_ROOTS
    return;
}

/*========================================================================
 * Function:        ClassType_SuperClass()
 * Overview:        returns the id of the super class
 *                  
 * Interface:
 *  parameters:     input pointer, output pointer, isBreak bool
 *  returns:        void
 *
 * Notes:           returns null if class is Object
 *=======================================================================*/

void ClassType_SuperClass(PacketInputStream *in, PacketOutputStream *out)
{
    unsigned long classID;
    INSTANCE_CLASS clazz;
    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    classID = inStream_readLong(in);
    clazz = (INSTANCE_CLASS)getObjectPtr(classID);

    outStream_writeClassID(out, (clazz == NULL || clazz->superClass == NULL ||
        ((clazz->clazz.accessFlags & ACC_INTERFACE) == ACC_INTERFACE)) ?
        NULL : (CLASS)(GET_CLASS_DEBUGGERID(&clazz->superClass->clazz)));
    END_TEMPORARY_ROOTS
 
}

/*========================================================================
 * Function:        ClassType_SetValues()
 * Overview:        set the value of one of an static variable
 * Interface:
 *   parameters:    pointer to debug header
 *   returns:
 *
 * Notes:       Packet format:
 *              Input:
 *                  object ID   Long
 *                  number of slots     Long
 *                  values  Long
 *
 *              Output:
 *                  debug header -      11 bytes
 *=======================================================================*/

static void ClassType_SetValues(PacketInputStream *in, PacketOutputStream *out)
{
    INSTANCE_CLASS clazz, fieldClazz;
    unsigned long classID, fieldClassID;
    long fieldID;
    FIELD fld;
    long value, value2;
    int i;
    long num;
    char *sig;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    IS_TEMPORARY_ROOT(sig, mallocBytes(DEBUG_MAX_SIG));
    classID = inStream_readLong(in);
    clazz = (INSTANCE_CLASS)getObjectPtr(classID);
    num = inStream_readLong(in);
    for (i = 0; i < num; i++) {
        fieldClassID = inStream_readLong(in);
        fieldID = inStream_readLong(in);
        IS_TEMPORARY_ROOT(fieldClazz, (INSTANCE_CLASS)getObjectPtr(fieldClassID));
        if (fieldClazz == NULL || fieldClazz->fieldTable == NULL) {
            outStream_setError(out, JDWP_Error_INVALID_FIELDID);
            goto breakout;
        }
        if (fieldID >= fieldClazz->fieldTable->length) {
            outStream_setError(out, JDWP_Error_INVALID_FIELDID);
            goto breakout;
        }
        START_TEMPORARY_ROOTS
        fld = getFieldAddress(fieldClazz, fieldID);
        if (inCurrentHeap(fld))
            IS_TEMPORARY_ROOT(fld, fld);
        change_Key_to_FieldSignature_inBuffer(
            fld->nameTypeKey.nt.typeKey, sig);
        switch(*sig) {
        case JDWP_Tag_BYTE:
        case JDWP_Tag_BOOLEAN:
            *(cell *)fld->u.staticAddress =
                inStream_readByte(in);
            break;
        case JDWP_Tag_CHAR:
        case JDWP_Tag_SHORT:
            *(cell *)fld->u.staticAddress =
                inStream_readShort(in);
            break;
        case JDWP_Tag_LONG:
        case JDWP_Tag_DOUBLE:
            value = inStream_readLong(in);
            value2 = inStream_readLong(in);
#if BIG_ENDIAN
            *(cell *)fld->u.staticAddress = value;
            *(&(((cell *)(fld->u.staticAddress))[1])) =
                value2;
#else
            *(cell *)fld->u.staticAddress = value2;
            *(&(((cell *)(fld->u.staticAddress))[1])) = value;
#endif
            break;
        case JDWP_Tag_OBJECT:
            *(cell *)fld->u.staticAddress =
            (cell)getObjectPtr(inStream_readLong(in));
            break;
        default:
            *(cell *)fld->u.staticAddress =
                inStream_readLong(in);
            break;
        }
        END_TEMPORARY_ROOTS
    }
breakout:
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        ObjectRefernce_ReferenceType
 * Overview:        returns the type tag of the object and what
 *                  class it is represented by
 *                  
 * Interface:
 *  parameters:     input pointer, output pointer, isBreak bool
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void ObjectReference_ReferenceType(PacketInputStream *in, PacketOutputStream *out)
{
    long objectID;
    OBJECT objectP;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    objectID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(objectP, getObjectPtr(objectID));
    
    /* we need to figure out what kind of object this reference points to */
    
    /* order matters here */
    if (objectID == ONLY_THREADGROUP_ID)    {
        outStream_writeByte(out, JDWP_TypeTag_CLASS);
        outStream_writeObjectRef(out,
            (OBJECT)GET_CLASS_DEBUGGERID(&JavaLangThread->clazz));
    }
    else if (validThread((THREAD)objectP)) {
        /* with the fix in outStream_writeObjectValue, we should never get here */
        outStream_writeByte(out,
            getJDWPTagType((CLASS)&JavaLangThread->clazz));
        outStream_writeLong(out,
            GET_CLASS_DEBUGGERID(&JavaLangThread->clazz));
    } 
    else if (validObject((OBJECT)objectP)) {
        outStream_writeByte(out, getJDWPTagType(objectP->ofClass));
        outStream_writeLong(out, GET_CLASS_DEBUGGERID(((OBJECT)(objectP))->ofClass));
    } else {
        outStream_setError(out, JDWP_Error_INVALID_OBJECT);
   }
         
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        ObjectReference_GetValues
 * Overview:        get the value of one of an instance variable
 * Interface:
 *   parameters:    pointer to debug header
 *   returns:
 *
 * Notes:       Packet format:
 *              Input:
 *                  object ID   long
 *                  number of slots     long
 *                  
 *              Output:
 *                  debug header -      11 bytes
 *                  count - long
 *                  values - long
 *=======================================================================*/
 
static void ObjectReference_GetValues(PacketInputStream *in, PacketOutputStream *out)
{
    long objectID;
    long num;
    FIELD fld = NULL;
    int i;
    char typeTag;
    long fieldID, fieldClassID;
    long offset;
    INSTANCE objectP;
    INSTANCE_CLASS fieldClazz;
    char *sig;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    objectID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(sig, mallocBytes(DEBUG_MAX_SIG));
    IS_TEMPORARY_ROOT(objectP, getObjectPtr(objectID));
    num = inStream_readLong(in);
    if (objectP == NULL) {
        outStream_writeLong(out, 0);
        goto breakout;
    }
    outStream_writeLong(out, num);
    for (i = 0; i < num; i++) {
        fieldClassID = inStream_readLong(in);
        fieldID = inStream_readLong(in);
        IS_TEMPORARY_ROOT(fieldClazz, (INSTANCE_CLASS)getObjectPtr(fieldClassID));
        if (fieldClazz == NULL || fieldClazz->fieldTable == NULL) {
            outStream_setError(out, JDWP_Error_INVALID_FIELDID);
            goto breakout;
        }
        if (fieldID >= fieldClazz->fieldTable->length) {
            outStream_setError(out, JDWP_Error_INVALID_FIELDID);
            goto breakout;
        }
        fld = getFieldAddress(fieldClazz, fieldID);
        START_TEMPORARY_ROOTS
        if (inCurrentHeap(fld))
            IS_TEMPORARY_ROOT(fld, fld);
        change_Key_to_FieldSignature_inBuffer(
            fld->nameTypeKey.nt.typeKey, sig);
        typeTag = *sig;
        if (fld->accessFlags & ACC_STATIC) {
            getStaticValue(out, fld, typeTag);
        } else {
            offset = fld->u.offset;
            if (fld->accessFlags & ACC_DOUBLE) {
                outStream_writeValue(out, typeTag,
                    (void *)(&objectP->data[offset].cell),
                    (void *)(&objectP->data[offset+1]), 1);
            } else {
                if (isObjectTag(typeTag))
                    outStream_writeObjectValue(out, (OBJECT)objectP->data[offset].cellp, typeTag);
                else
                    outStream_writeValue(out, typeTag,
                        (void *)(&objectP->data[offset].cell), 0, 1);
            }
        }
        END_TEMPORARY_ROOTS
    }
breakout:
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        ObjectReference_SetValues
 * Overview:        set the value of one of an instance variable
 * Interface:
 *   parameters:    pointer to debug header
 *   returns:
 *
 * Notes:       Packet format:
 *              Input:
 *                  object ID   long
 *                  number of slots long
 *                  values  long
 
 *              Output:
 *                  debug header -      11 bytes
 *=======================================================================*/
 
static void ObjectReference_SetValues(PacketInputStream *in, PacketOutputStream *out)
{
    long objectID;
    long fieldID, fieldClassID;
    FIELD fld;
    long value, value2;
    int i;
    long num, offset;
    char *sig;
    INSTANCE objectP;
    INSTANCE_CLASS fieldClazz;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    objectID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(sig, mallocBytes(DEBUG_MAX_SIG));
    IS_TEMPORARY_ROOT(objectP, getObjectPtr(objectID));
    if (objectP == NULL) {
        outStream_writeLong(out, 0);
        goto breakout;
    }
    num = inStream_readLong(in);
    for (i = 0; i < num; i++) {
        fieldClassID = inStream_readLong(in);
        fieldID = inStream_readLong(in);

        IS_TEMPORARY_ROOT(fieldClazz, (INSTANCE_CLASS)getObjectPtr(fieldClassID));
        if (fieldClazz == NULL || fieldClazz->fieldTable == NULL) {
            outStream_setError(out, JDWP_Error_INVALID_FIELDID);
            goto breakout;
        }
        if (fieldID >= fieldClazz->fieldTable->length) {
            outStream_setError(out, JDWP_Error_INVALID_FIELDID);
            goto breakout;
        }
        fld = getFieldAddress(fieldClazz, fieldID);
        START_TEMPORARY_ROOTS
        if (inCurrentHeap(fld))
            IS_TEMPORARY_ROOT(fld, fld);
        change_Key_to_FieldSignature_inBuffer(
            fld->nameTypeKey.nt.typeKey, sig);
        offset = fld->u.offset;
        switch(*sig) {
        case JDWP_Tag_BYTE:
        case JDWP_Tag_BOOLEAN:
            objectP->data[offset].cell = inStream_readByte(in);
            break;
        case JDWP_Tag_CHAR:
        case JDWP_Tag_SHORT:
            objectP->data[offset].cell = inStream_readShort(in);
            break;
        case JDWP_Tag_LONG:
        case JDWP_Tag_DOUBLE:
            value = inStream_readLong(in);
            value2 = inStream_readLong(in);
#if BIG_ENDIAN
            objectP->data[offset].cell = value;
            objectP->data[offset+1].cell = value2;
#else
            objectP->data[offset].cell = value2;
            objectP->data[offset+1].cell = value;
#endif
            break;
        case JDWP_Tag_OBJECT:
            objectP->data[offset].cell =
                (cell)getObjectPtr(inStream_readLong(in));
            break;
        default:
            objectP->data[offset].cell = inStream_readLong(in);
            break;
        }
        END_TEMPORARY_ROOTS
    }
breakout:
    END_TEMPORARY_ROOTS
}

void StringReference_Value(PacketInputStream *in, PacketOutputStream *out)
{
    STRING_INSTANCE str;
    unsigned long stringID = inStream_readLong(in);
    char *buf;
    int buflen;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    IS_TEMPORARY_ROOT(str, (STRING_INSTANCE)getObjectPtr(stringID));
    buflen = str->length + 1;
    IS_TEMPORARY_ROOT(buf, mallocBytes(buflen));

    getStringContentsSafely(str, buf, buflen);
    outStream_writeString(out, (CONST_CHAR_HANDLE)&buf);
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        ThreadReference_Name()
 * Overview:        send the specified thread's name
 * Interface:
 *   parameters:    pointer to debug header
 *   returns:       none
 *                  
 *
 * Notes:
 *=======================================================================*/

void ThreadReference_Name(PacketInputStream *in, PacketOutputStream *out)
{
    long threadID;
    THREAD thread;
    char *nullString = "";

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    threadID = inStream_readLong(in);
 
    IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(threadID));

    if (validThread(thread)) {
        DECLARE_TEMPORARY_ROOT(char *, name, mallocBytes(10 + sizeof(long)*2 + 1));
        
        if (thread == MainThread)
            sprintf(name, "KVM_main");
        else
            sprintf(name, "KVM_Thread%lx", (long)thread);

        outStream_writeString(out, (CONST_CHAR_HANDLE)&name);
    }
    else {
        outStream_writeString(out, (CONST_CHAR_HANDLE)&nullString);
    }
    END_TEMPORARY_ROOTS
    
}

/*========================================================================
 * Function:        ThreadReference_Suspend()
 * Overview:        suspends a thread
 *                  
 * Interface:
 *  parameters:     input pointer, output pointer, isBreak bool
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void ThreadReference_Suspend(PacketInputStream *in, PacketOutputStream *out)
{
    long threadID;
    THREAD thread;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    threadID = inStream_readLong(in);
  
    IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(threadID));

    if (validThread(thread)) {
        suspendSpecificThread(thread);
    }
    else {
        /* outStream_setError(out, JDWP_Error_INVALID_THREAD); */
    }
    END_TEMPORARY_ROOTS

}

/*========================================================================
 * Function:        ThreadReference_Resume
 * Overview:        resumes a thread
 *                  
 * Interface:
 *  parameters:     input pointer, output pointer, isBreak bool
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void ThreadReference_Resume(PacketInputStream *in, PacketOutputStream *out)
{
    long threadID;
    THREAD thread;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    threadID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(threadID));

    if (validThread(thread)) {
        resumeSpecificThread(thread);
    } else {
        outStream_setError(out, JDWP_Error_INVALID_THREAD);
    }
    allThreadsSuspended = FALSE;
    setResumeOK(TRUE);
    END_TEMPORARY_ROOTS

}

/*========================================================================
 * Function:        ThreadReference_Status()
 * Overview:        returns the JDWP thread status
 *                  
 * Interface:
 *  parameters:     input pointer, output pointer, isBreak bool
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void ThreadReference_Status(PacketInputStream *in, PacketOutputStream *out)
{
    long threadID;
    THREAD thread;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    threadID = inStream_readLong(in);

    IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(threadID));

    if (validThread(thread)) {
        outStream_writeLong(out, getJDWPThreadStatus(thread));
        outStream_writeLong(out, getJDWPThreadSuspendStatus(thread));
    } else {
        /* outStream_setError(out, JDWP_Error_INVALID_THREAD); */
        outStream_writeLong(out, JDWP_ThreadStatus_ZOMBIE);
        outStream_writeLong(out, !JDWP_SuspendStatus_SUSPEND_STATUS_SUSPENDED);
    }
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        ThreadReference_SuspendCount()
 * Overview:        returns the number of times this thread 
 *                  has been suspended
 *                  
 * Interface:
 *  parameters:     input pointer, output pointer, isBreak bool
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void ThreadReference_SuspendCount(PacketInputStream *in, PacketOutputStream *out)
{
    long threadID;
    THREAD thread;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    threadID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(threadID));

    if (validThread(thread)) {
        outStream_writeLong(out, thread->debugSuspendCount);
    }
    else {
        outStream_setError(out, JDWP_Error_INVALID_THREAD);
    }
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        ThreadReference_Frames()
 * Overview:        dump all frames or a particular frame
 * Interface:
 *   parameters:    debug header pointer
 *   returns:       nothing
 *
 * Notes:       Packet format:
 *              Input:
 *                  thread ID   long
 *                  startIndex  long
 *                  len         long
 *              Output:
 *                  debug header    11 bytes
 *                  frame count-    long
 *                  For each frame:
 *                      current fp long
 *                      location    17 bytes
 *=======================================================================*/

static void ThreadReference_Frames(PacketInputStream *in, PacketOutputStream *out)
{
    FRAME currentFP;
    BYTE *currentIP;
    long framecount = 0;
    METHOD methodID;
    INSTANCE_CLASS clazz;
    long startIndex, len, index;
    long threadID;
    THREAD thread;


    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    threadID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(threadID));

    if (!validThread(thread)) {
        outStream_setError(out, JDWP_Error_INVALID_THREAD);
        goto breakout;
    }

    startIndex = inStream_readLong(in);
    len = inStream_readLong(in);
    if (CurrentThread)
        storeExecutionEnvironment(CurrentThread);

    currentFP = thread->fpStore;
    currentIP = thread->ipStore;
    
    /* first get the count of frames on the stack */
    while (currentFP != NIL) {
        framecount++;
        currentFP = currentFP->previousFp;
    }
    if (startIndex >= framecount) {
        outStream_setError(out, JDWP_Error_INVALID_INDEX);
        goto breakout;
    }

    if (len == -1) {
        framecount = framecount - startIndex;
    } else {
        if (len < (framecount - startIndex))
            framecount = len;
        else
            framecount = framecount - startIndex;
    }
    outStream_writeLong(out, framecount);
    while (framecount-- > 0) {
        /* may have had a GC so we need to find FP again */
        currentFP = thread->fpStore;
        currentIP = thread->ipStore;
        index = 0;
        while (startIndex > index) {
            currentIP = currentFP->previousIp;
            currentFP = currentFP->previousFp;
            index++;
        }

        if (currentFP != NIL) {
            outStream_writeLong(out, index);
            methodID = currentFP->thisMethod;
            clazz = methodID->ofClass;

            outStream_writeLocation(out, JDWP_TypeTag_CLASS,
                (CLASS)(GET_CLASS_DEBUGGERID(&clazz->clazz)),
                (METHOD)getMethodIndex(clazz, (METHOD)methodID),
                (unsigned long)(currentIP - methodID->u.java.code));
            startIndex++;
        }
    }
breakout:
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        ThreadReference_FrameCount()
 * Overview:        returns the frame depth of a thread
 *                  
 * Interface:
 *  parameters:     input pointer, output pointer, isBreak bool
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void ThreadReference_FrameCount(PacketInputStream *in, PacketOutputStream *out)
{
    long framecount;
    FRAME currentFP;
    long threadID;
    THREAD thread;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    threadID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(threadID));

    if (validThread(thread)) {
        if (thread->state & THREAD_DBG_SUSPENDED) {

            currentFP = thread->fpStore;
            framecount = 0;
            while (currentFP != NIL) {
                framecount++;
                currentFP = currentFP->previousFp;
            }
            outStream_writeLong(out, framecount);
        } else {
            outStream_writeLong(out, JDWP_Error_THREAD_NOT_SUSPENDED);
        }
    } else {
        outStream_setError(out, JDWP_Error_INVALID_THREAD);
    }
    END_TEMPORARY_ROOTS
}

void ArrayReference_Length(PacketInputStream *in, PacketOutputStream *out)
{
    ARRAY array;
    unsigned long arrayID;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    arrayID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(array, (ARRAY)getObjectPtr(arrayID));

    if (array && IS_ARRAY_CLASS(array->ofClass)) {
        outStream_writeLong(out, array->length);
    } else {    
        outStream_writeLong(out, 0);
    }
    END_TEMPORARY_ROOTS
}


void ArrayReference_SetValues(PacketInputStream *in, PacketOutputStream *out)
{
    unsigned long arrayID;
    long firstIndex, length, i;
    BYTE tag;
    long value, value2;
    ARRAY array;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    arrayID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(array, (ARRAY)getObjectPtr(arrayID));
    tag = (BYTE)((array->ofClass->gcType == GCT_ARRAY) 
        ? typeCodeToSignature((char)(array->ofClass->u.primType))
        : (BYTE)JDWP_Tag_OBJECT);
    if (array) {
        firstIndex = inStream_readLong(in);
        length = inStream_readLong(in);
        if (tag == JDWP_Tag_LONG || tag == JDWP_Tag_DOUBLE) {
            for (i=firstIndex*2; i < (firstIndex+length)*2; i+=2) {
                value = inStream_readLong(in);
            value2 = inStream_readLong(in);
#if BIG_ENDIAN
            array->data[i].cell = value;
            array->data[i+1].cell = value2;
#else
            array->data[i].cell = value2;
            array->data[i+1].cell = value;
#endif
        }
        } else {
            for (i=firstIndex; i < firstIndex+length; i++) {
            switch(tag) {
            case JDWP_Tag_BYTE:
            case JDWP_Tag_BOOLEAN:
            ((BYTEARRAY)array)->bdata[i] = inStream_readByte(in);
            break;
            case JDWP_Tag_CHAR:
            case JDWP_Tag_SHORT:
            ((SHORTARRAY)array)->sdata[i] = inStream_readShort(in);
            break;
            case JDWP_Tag_OBJECT:
            array->data[i].cell = (cell)getObjectPtr(inStream_readLong(in));
                break;
            default:
            value = inStream_readLong(in);
            array->data[i].cell = value;
            }
        }
        }
    }
    END_TEMPORARY_ROOTS
}

void ArrayReference_GetValues(PacketInputStream *in, PacketOutputStream *out)
{
    unsigned long arrayID;
    long firstIndex, length, i;
    BYTE tag;
    long value;
    ARRAY array;
    char *buffer;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    arrayID = inStream_readLong(in);
    IS_TEMPORARY_ROOT(array, (ARRAY)getObjectPtr(arrayID));
    if (array) {
        firstIndex = inStream_readLong(in);
        length = inStream_readLong(in);

        if (length == -1)
        length = array->length;
        IS_TEMPORARY_ROOT(buffer, mallocBytes(256));
        buildClassSignature((CLASS)array->ofClass, (CONST_CHAR_HANDLE)&buffer);
        tag = buffer[1];

        outStream_writeByte(out, tag);  /* type of array */
        outStream_writeLong(out, length);    /* number of components being sent */

        if (tag == JDWP_Tag_LONG || tag == JDWP_Tag_DOUBLE) {
            for (i=firstIndex*2; i < (firstIndex+length)*2; i+=2)
                outStream_writeValue(out, tag, &array->data[i].cell,
                &array->data[ i+1 ].cell, 0);
        } else {
            for (i=firstIndex; i < firstIndex+length; i++) {
            switch (tag) {
            case JDWP_Tag_BYTE:
            case JDWP_Tag_BOOLEAN:
            value = ((BYTEARRAY)array)->bdata[i];
                outStream_writeValue(out, tag,
                &value, 0, 0);
            break;
            case JDWP_Tag_SHORT:
            case JDWP_Tag_CHAR:
            value = ((SHORTARRAY)array)->sdata[i];
                outStream_writeValue(out, tag,
                &value, 0, 0);
            break;
            default:
                if (isObjectTag(tag))
                    outStream_writeObjectValue(out, (OBJECT)array->data[i].cellp, tag);
                else
                    outStream_writeValue(out, tag,
                        &array->data[i].cell, 0, 0);
            break;
            }
        }
        }
    }
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        createEvent
 * Overview:        creates a new event
 *                  
 * Interface:
 *  parameters:     
 *  returns:        void
 *
 * Notes:           since there is only one thread group all threads
 *                  are its children
 *=======================================================================*/

static void createEvent(PacketInputStream *in, PacketOutputStream *out, BYTE kind)
{
    VMEventPtr ep, tep;
    int i;
    BYTE modKind;
    EVENTMODIFIER curModSlot;
    EVENTMODIFIER newMod;
    BYTE alreadyExists;
    THREAD thread;
    char *tmpName;
    
    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    ep = createVmEvent();
    
    ep->kind = kind;
    ep->suspendPolicy = inStream_readByte(in);
    ep->numModifiers = inStream_readLong(in);
    ep->count = 0;
    ep->count_active = FALSE;

    alreadyExists = FALSE;
    for (i=0; i < ep->numModifiers && !alreadyExists; i++) {

        modKind = inStream_readByte(in);

        newMod = NIL;        

        switch(modKind) {
        case JDWP_EventRequest_Set_Out_modifiers_Modifier_Count:
            ep->count = inStream_readLong(in);
        if (ep->count > 0)
            ep->count_active = TRUE;
            break;
        case JDWP_EventRequest_Set_Out_modifiers_Modifier_Conditional:
            inStream_readLong(in); 
            break;
        case JDWP_EventRequest_Set_Out_modifiers_Modifier_ThreadOnly:
            inStream_readLong(in); 
            break;
        case JDWP_EventRequest_Set_Out_modifiers_Modifier_ClassOnly:
            newMod = (EVENTMODIFIER)getModPtr();
            newMod->kind = modKind;
            newMod->u.classID = inStream_readLong(in);
            break;
        case JDWP_EventRequest_Set_Out_modifiers_Modifier_ClassMatch:
            newMod = (EVENTMODIFIER)getModPtr();
            newMod->kind = modKind;
            IS_TEMPORARY_ROOT(tmpName, inStream_readString(in));
        newMod->u.classMatch.className =
        (char *)callocPermanentObject(ByteSizeToCellSize((strlen(tmpName) + 1)));
        strcpy(newMod->u.classMatch.className, tmpName);
            break;
        case JDWP_EventRequest_Set_Out_modifiers_Modifier_ClassExclude:
            inStream_readString(in); 
            break;
        case JDWP_EventRequest_Set_Out_modifiers_Modifier_LocationOnly:

            switch(kind) {
            case JDWP_EventKind_BREAKPOINT:
            case JDWP_EventKind_FIELD_ACCESS:
            case JDWP_EventKind_FIELD_MODIFICATION:
            case JDWP_EventKind_SINGLE_STEP:
            case JDWP_EventKind_EXCEPTION:
                newMod = (EVENTMODIFIER)getModPtr();
                newMod->kind = modKind;
                newMod->u.loc.tag = inStream_readByte(in);
                newMod->u.loc.classID = inStream_readLong(in);
                if (newMod->u.loc.classID == 0) {
                    outStream_setError(out, JDWP_Error_ILLEGAL_ARGUMENT);
                    return;        /* skunked! */
                }
                newMod->u.loc.methodIndex =
                    (unsigned long)inStream_readMethodID(in);
                inStream_readLong64(in, &newMod->u.loc.offset, &dummyLong);
                if ((tep = installBreakpoint(in, out, &newMod)) != NIL) {
                    ep->inUse = FALSE;
                    goto breakout;
                }
                break;
            }

            break;
        case JDWP_EventRequest_Set_Out_modifiers_Modifier_ExceptionOnly:
            switch(kind) {
            case JDWP_EventKind_EXCEPTION:
                newMod = (EVENTMODIFIER)getModPtr();
                newMod->kind = modKind;
                newMod->u.exp.classID = inStream_readLong(in);
                newMod->u.exp.sig_caught = inStream_readBoolean(in);
                newMod->u.exp.sig_uncaught = inStream_readBoolean(in);
                break;
            }
            break;
        case JDWP_EventRequest_Set_Out_modifiers_Modifier_FieldOnly:
            inStream_readObjectRef(in);
            inStream_readFieldID(in);
            break;
        case JDWP_EventRequest_Set_Out_modifiers_Modifier_Step:
            switch (kind) {
            case JDWP_EventKind_SINGLE_STEP:

                newMod = (EVENTMODIFIER)getModPtr();
                newMod->kind = modKind;
                newMod->u.step.threadID = inStream_readLong(in);
                thread = GET_VMTHREAD(newMod->u.step.threadID);
                if (thread == NULL || (thread->state & THREAD_DEAD)) {
                    /*
                     * If you try to single step after suspending because
                     * of an uncaught exception event we'll get sent the
                     * thread id of the thread that had the exception.
                     * That thread is dead though.
                     */
                    outStream_setError(out, JDWP_Error_INVALID_THREAD);
                    return;
                }

                newMod->u.step.size = inStream_readLong(in);
                newMod->u.step.depth = inStream_readLong(in);
                
                /* query the _proxy_ for next line location */                        
        {
        CEModPtr cep = GetCEModifier();

                setEvent_NeedSteppingInfo(newMod->u.step.threadID, cep);
        /*
         * At this point, the proxy has made sure no more commands
         * from the debugger get sent until we get our line number
         * info from the proxy
         */
        processBreakCommands();
        /*
         * Once we get back here, then we know we've gotten the
         * command from the proxy with the line number info put into
         * the cep structure.
         */
 
                /* setup the relevant info */
        thread = GET_VMTHREAD(newMod->u.step.threadID);
                thread->isStepping = TRUE;
/*
                newMod->u.step.thread->stepInfo.target.cl =
            cep->step.target.cl;
                newMod->u.step.thread->stepInfo.target.method =
            cep->step.target.method;
*/
                thread->stepInfo.target.offset = cep->step.target.offset;
                thread->stepInfo.size = newMod->u.step.size;
                thread->stepInfo.depth = newMod->u.step.depth;
                thread->stepInfo.fp = thread->fpStore;
                thread->stepInfo.offset =
            ip_global - thread->fpStore->thisMethod->u.java.code;
                thread->stepInfo.nextLineOffs = cep->step.nextLineOffs;
        FreeCEModifier(cep);
        }

                break;
            }
            break;
        }

        if (newMod != NIL) {
            /* insert the mod */
            newMod->next = NIL;
            if (ep->mods == NIL) {
                ep->mods = newMod;
                curModSlot = ep->mods;
            }
            else {
                curModSlot->next = newMod;
                curModSlot = curModSlot->next;
            }
        }
    }
breakout:    
    END_TEMPORARY_ROOTS
    if (alreadyExists == TRUE)
        outStream_writeObjectRef(out, (OBJECT)tep);
    else
        outStream_writeObjectRef(out, (OBJECT)ep);

}

/*========================================================================
 * Function:        EventRequest_Set()
 * Overview:        sets up an event to happen such as a breakpoint
 * Interface:
 *   parameters:    pointer to debug header
 *   returns:       nothing
 *
 * Notes:
 *=======================================================================*/

static void EventRequest_Set(PacketInputStream *in, PacketOutputStream *out)
{
    unsigned char eventType;
    long dbg_eventType;

    eventType = inStream_readByte(in);

    switch (eventType) {
    case JDWP_EventKind_BREAKPOINT:
        dbg_eventType = Dbg_EventKind_BREAKPOINT;
        break;
    case JDWP_EventKind_SINGLE_STEP:
        dbg_eventType = Dbg_EventKind_SINGLE_STEP;
        break;
    case JDWP_EventKind_THREAD_START:
        dbg_eventType = Dbg_EventKind_THREAD_START;
        break;
    case JDWP_EventKind_THREAD_END:
        dbg_eventType = Dbg_EventKind_THREAD_END;
        break;
    case JDWP_EventKind_CLASS_PREPARE:
        dbg_eventType = Dbg_EventKind_CLASS_PREPARE;
        break;
    case JDWP_EventKind_CLASS_LOAD:
        dbg_eventType = Dbg_EventKind_CLASS_LOAD;
        break;
    case JDWP_EventKind_CLASS_UNLOAD:
        dbg_eventType = Dbg_EventKind_CLASS_UNLOAD;
        break;
    case JDWP_EventKind_VM_INIT:
        dbg_eventType = Dbg_EventKind_VM_INIT;
        break;
    case JDWP_EventKind_VM_DEATH:
        dbg_eventType = Dbg_EventKind_VM_DEATH;
        break;
    case JDWP_EventKind_METHOD_EXIT:
        dbg_eventType = Dbg_EventKind_METHOD_EXIT;
        break;
    case JDWP_EventKind_EXCEPTION_CATCH:
        dbg_eventType = Dbg_EventKind_EXCEPTION_CATCH;
        break;
    case JDWP_EventKind_USER_DEFINED:
        dbg_eventType = Dbg_EventKind_USER_DEFINED;
        break;
    case JDWP_EventKind_METHOD_ENTRY:
        dbg_eventType = Dbg_EventKind_METHOD_ENTRY;
        break;
    case JDWP_EventKind_FIELD_MODIFICATION:
        dbg_eventType = Dbg_EventKind_FIELD_MODIFICATION;
        break;
    case JDWP_EventKind_FRAME_POP:
        dbg_eventType = Dbg_EventKind_FRAME_POP;
        break;
    case JDWP_EventKind_FIELD_ACCESS:
        dbg_eventType = Dbg_EventKind_FIELD_ACCESS;
        break;
    case JDWP_EventKind_EXCEPTION:
        dbg_eventType = Dbg_EventKind_EXCEPTION;
        break;
    default:
        outStream_setError(out, JDWP_Error_INVALID_EVENT_TYPE);
        return;
    }

    createEvent(in, out, eventType);
    
    addNotification(dbg_eventType);

}

/*========================================================================
 * Function:        clearEvent()
 * Overview:        clear one or all events
 * Interface:
 *   parameters:    pointer to debug header,  flag
 *   returns:       nothing
 *
 * Notes:       Packet format:
 *              Input:
 *                  event type  byte
 *                  event ID    int
 *              Output:
 *                  debug header -  11 bytes
 *=======================================================================*/

static void clearEvent(VMEventPtr ep)
{
    EVENTMODIFIER mod = NIL;
    INSTANCE_CLASS clazz;
    METHOD thisMethod;
    char *start;
    THREAD thread;


    START_TEMPORARY_ROOTS
    switch (ep->kind) {
    case JDWP_EventKind_BREAKPOINT:
        mod = getModifier(ep,
        JDWP_EventRequest_Set_Out_modifiers_Modifier_LocationOnly);
    IS_TEMPORARY_ROOT(thisMethod, getMethodFromIndex(mod->u.loc.classID,
        mod->u.loc.methodIndex));
        if (mod != NIL) {
            clazz = (INSTANCE_CLASS)getObjectPtr(mod->u.loc.classID);
            start = (char *)clazz->constPool;
            setBreakpoint(thisMethod, mod->u.loc.offset, start,
                mod->u.loc.opcode);
        }
        break;
    case JDWP_EventKind_SINGLE_STEP:
        mod = getModifier(ep,
        JDWP_EventRequest_Set_Out_modifiers_Modifier_Step);
        if (mod != NIL) {
            IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(mod->u.step.threadID));
            if (thread != NULL)
                thread->isStepping = FALSE;
        }
        break;
    }
    ep->active = FALSE;
    ep->inUse = FALSE;
    END_TEMPORARY_ROOTS
}

void clearAllBreakpoints()
{
    VMEventPtr ep;
    
    ep = eventHead;
    if (ep == NIL) return;
    
    do {
        if (ep->kind == JDWP_EventKind_BREAKPOINT)
            clearEvent(ep);
    } while ((ep = ep->next));
}

/*========================================================================
 * Function:        EventRequest_Clear()
 * Overview:        cancel a specific event request
 *                  
 * Interface:
 *  parameters:     input pointer, output pointer, isBreak bool
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

static void EventRequest_Clear(PacketInputStream *in, PacketOutputStream *out)
{
    BYTE eventType = inStream_readByte(in);
    VMEventPtr epid = (VMEventPtr)inStream_readObjectRef(in);
    VMEventPtr ep;
    
    eventType = 0;      /* satisfy GCC */
    ep = eventHead;
    if (ep == NIL) return;
    
    do {
        if (ep == epid)
            clearEvent(ep);
    } while ((ep = ep->next));

}

/*========================================================================
 * Function:        EventRequest_ClearAllBreakpoints()
 * Overview:        clears all breakpoint notifications
 *                  
 * Interface:
 *  parameters:     input pointer, output pointer, isBreak bool
 *  returns:        void
 *
 * Notes:
 *=======================================================================*/

void EventRequest_ClearAllBreakpoints(PacketInputStream *in, PacketOutputStream *out)
{
    clearAllBreakpoints();
}

void StackFrame_GetValues(PacketInputStream *in, PacketOutputStream *out)
{
    long threadID;
    THREAD thread;
    unsigned long frameID;
    FRAME currentFP;
    long numValues, i, slot;
    BYTE tag;
    cell *localP;
    unsigned int index = 0;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    threadID = inStream_readLong(in);
    frameID = inStream_readLong(in);
    numValues = inStream_readLong(in);

    if (CurrentThread)
        storeExecutionEnvironment(CurrentThread);
    IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(threadID));
    outStream_writeLong(out, numValues);
    for (i=0; i < numValues; i++) {
        currentFP = thread->fpStore;
        while (frameID > index) {
            currentFP = currentFP->previousFp;
            index++;
        }
        localP = FRAMELOCALS(currentFP);
        slot = inStream_readLong(in);
        tag  = inStream_readByte(in);
        if (tag == JDWP_Tag_LONG || tag == JDWP_Tag_DOUBLE)
            outStream_writeValue(out, tag, (void *)(localP + slot),
                (void *)(localP + slot + 1), 1);
        else {
            if (isObjectTag(tag))
                outStream_writeObjectValue(out, (OBJECT)*(OBJECT *)(localP + slot), tag);
            else
                outStream_writeValue(out, tag, (void *)(localP+slot),
                    0, 1);
        }
    }
    END_TEMPORARY_ROOTS
}

/*========================================================================
 * Function:        StackFrame_SetValues
 * Overview:        set the value of one of the locals on the stack
 * Interface:
 *   parameters:    pointer to debug header
 *   returns:
 *
 * Notes:       Packet format:
 *              Input:
 *                  thread ID   Long
 *                  frame ID    Long
 *                  number of slots Long
 *                  
 *              Output:
 *                  debug header -      11 bytes
 *=======================================================================*/

static void StackFrame_SetValues(PacketInputStream *in, PacketOutputStream *out)
{
    long threadID;
    THREAD thread;
    unsigned long frameID;
    FRAME currentFP;
    long numslots, value, value2;
    unsigned char type;
    cell *localP;
    int i;
    unsigned long index = 0;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    threadID = inStream_readLong(in);
    frameID = inStream_readLong(in);

    if (CurrentThread)
        storeExecutionEnvironment(CurrentThread);
    IS_TEMPORARY_ROOT(thread, GET_VMTHREAD(threadID));
    currentFP = thread->fpStore;
    while (frameID > index) {
        currentFP = currentFP->previousFp;
        index++;
    }
    localP = FRAMELOCALS(currentFP);
    numslots = inStream_readLong(in);
    for (i = 0; i < numslots; i++) {
        index = inStream_readLong(in);
        type = inStream_readByte(in);
        if (type == JDWP_Tag_LONG || type == JDWP_Tag_DOUBLE) {
            value = inStream_readLong(in);
            value2 = inStream_readLong(in);
#if BIG_ENDIAN
            localP[index] = value;
            localP[index+1] = value2;
#else
            localP[index] = value2;
            localP[index+1] = value;
#endif
        } else {
            switch(type) {
            case JDWP_Tag_BYTE:
            case JDWP_Tag_BOOLEAN:
                localP[index] = inStream_readByte(in);
                break;
            case JDWP_Tag_SHORT:
                case JDWP_Tag_CHAR:
                localP[index] = inStream_readShort(in);
                break;
            case JDWP_Tag_OBJECT:
                localP[index] = (cell)getObjectPtr(inStream_readLong(in));
                break;
            default:
                localP[index] = inStream_readLong(in);
            }
            }
    }
    END_TEMPORARY_ROOTS
    return;
}

void KVMCommands_HandShake(PacketInputStream *in, PacketOutputStream *out) {

    char *versionString;
    int major;
    int minor;
    char *kvmString = "KVM Reference Debugger Interface";

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(out, out);
    IS_TEMPORARY_ROOT(in, in);
    IS_TEMPORARY_ROOT(versionString, inStream_readString(in));
    major = inStream_readByte(in);
    minor = inStream_readByte(in);
    outStream_writeString(out, (CONST_CHAR_HANDLE)&kvmString);
    outStream_writeLong(out, 0x7FFF);  /* set appropriate bits */
    setResumeOK(TRUE);
    END_TEMPORARY_ROOTS
}

void KVMCommands_GetSteppingInfo(PacketInputStream *in, PacketOutputStream *out)
{

    CEModPtr cep = (CEModPtr)inStream_readLong(in);
/*
    cep->step.target.cl = 0x00;
    cep->step.target.method = 0x00;
*/
    inStream_readLong64(in, &cep->step.target.offset, &dummyLong);
    inStream_readLong64(in, &cep->step.nextLineOffs, &dummyLong);
    setResumeOK(TRUE);
}

/*========================================================================
 * Function:        InitDebugger
 * Type:            private function
 * Overview:        Open listener socket to listen for debugger commands
 *
 * Interface:
 *   parameters:    <none>
 *   returns:       true - no error
 *                  false -  Error
 * Note:
 *=======================================================================*/

bool_t InitDebugger()
{

    if (!dbgInitialized()) {
        InitDebuggerIO();
        if (GetDebugger(6000) == FALSE)
            return FALSE;
    }
    requestID = 1;

    funcArray = (void **)callocPermanentObject(((JDWP_HIGHEST_COMMAND_SET+1) * sizeof (void *)) >> log2CELL);
    funcArray[JDWP_COMMAND_SET(VirtualMachine)] = (void *)VirtualMachine_Cmds;
    funcArray[JDWP_COMMAND_SET(ReferenceType)] = (void *)ReferenceType_Cmds;
    funcArray[JDWP_COMMAND_SET(ClassType)] = (void *)ClassType_Cmds;
    funcArray[JDWP_COMMAND_SET(ArrayType)] = (void *)ArrayType_Cmds;
    funcArray[JDWP_COMMAND_SET(Field)] = (void *)Field_Cmds;
    funcArray[JDWP_COMMAND_SET(Method)] = (void *)Method_Cmds;
    funcArray[JDWP_COMMAND_SET(ObjectReference)] = (void *)ObjectReference_Cmds;
    funcArray[JDWP_COMMAND_SET(StringReference)] = (void *)StringReference_Cmds;
    funcArray[JDWP_COMMAND_SET(ThreadReference)] = (void *)ThreadReference_Cmds;
    funcArray[JDWP_COMMAND_SET(ThreadGroupReference)] = (void *)ThreadGroupReference_Cmds;
    funcArray[JDWP_COMMAND_SET(ClassLoaderReference)] = (void *)ClassLoaderReference_Cmds;
    funcArray[JDWP_COMMAND_SET(ArrayReference)] = (void *)ArrayReference_Cmds;
    funcArray[JDWP_COMMAND_SET(EventRequest)] = (void *)EventRequest_Cmds;
    funcArray[JDWP_COMMAND_SET(StackFrame)] = (void *)StackFrame_Cmds;
    funcArray[JDWP_COMMAND_SET(ClassObjectReference)] = (void *)ClassObjectReference_Cmds;

    KVMFuncArray = (void **)callocPermanentObject((2 * sizeof (void *)) >> log2CELL);

    KVMFuncArray[0] = (void *)KVMCommands;

    debugRoot = (POINTERLIST)callocObject(SIZEOF_POINTERLIST(OBJECT_ROOT_SIZE),
        GCT_WEAK_POINTERLIST);
    debugRoot->length = (OBJECT_ROOT_SIZE);
    makeGlobalRoot((cell **)&debugRoot);
    bucketMap = (unsigned long *)mallocBytes(OBJECT_ROOT_SIZE * CELL);
    memset(bucketMap, 0, (OBJECT_ROOT_SIZE * CELL));
    makeGlobalRoot((cell **)&bucketMap);
    createHashTable(&debuggerHashTable, DEBUGGER_HASH_SIZE);

    vmDebugReady = TRUE;
    /*
     * Get the handshake from the proxy
     */
    processBreakCommands();
    return TRUE;
}

/*========================================================================
 * Function:        GetDebugger()
 * Overview:        see if a debugger is connecting to our listener socket
 * Interface:
 *   parameters:    timeout
 *   returns:       true - OK, false - bad
 *
 * Notes:
 *=======================================================================*/

bool_t GetDebugger(int timeout)
{
    bool_t retval;

    retval = GetDebuggerChannel(timeout);
    return retval;
}

/*========================================================================
 * Function:        CloseDebuggerIO()
 * Overview:        close the sockets
 * Interface:
 *   parameters:    none
 *   returns:       none
 *
 * Notes:
 *=======================================================================*/

void CloseDebugger()
{
    CloseDebuggerIO();
}

/*========================================================================
 * Function:        processDebugCmds()
 * Overview:        parse commands from debugger
 * Interface:
 *   parameters:    select timeout in milliseconds, -1 = forever
 *   returns:       nothing
 *
 * Notes:
 *=======================================================================*/

void ProcessDebugCmds(int timeout)

{
    struct CmdPacket *cmd;
    PacketInputStream  *in;
    PacketOutputStream  *out;
    void **f2Array;
    CommandHandler func;
        Packet *p;

    if (!vmDebugReady) {
        return;
    }
    if (!dbgCharAvail(allThreadsSuspended ? -1 : timeout))
        return;
    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(in, (PacketInputStream *)callocObject(SIZEOF_PACKETSTREAM,
        GCT_POINTERLIST));
    in->numPointers = 0;
    IS_TEMPORARY_ROOT(out,
        (PacketOutputStream *)callocObject(SIZEOF_PACKETSTREAM,
        GCT_POINTERLIST));
    in->packet = (Packet *)callocObject(SIZEOF_PACKET, GCT_POINTERLIST);
    in->numPointers = 1;
    in->segment = in->packet->type.cmd.data =
        (PacketData *)callocObject(SIZEOF_PACKETDATA, GCT_POINTERLIST);
    in->packet->type.cmd.numPointers = 1;
    in->numPointers = 2;
    in->packet->type.cmd.data->data =
        (unsigned char *)mallocBytes(MAX_INPUT_PACKET_SIZE);
    in->segment->numPointers = 1;
    if (dbgReceivePacket(in) == SYS_ERR)
        goto getOut;

    p = in->packet;
    if (p->type.cmd.flags & FLAGS_Reply) {
        /* It's a reply, just drop it */
        goto getOut;
    }
    cmd = &p->type.cmd;

    if (cmd->cmdSet == KVM_CMDSET) {
        f2Array = (void **)KVMFuncArray[0];
    } else {
        if (cmd->cmdSet > JDWP_HIGHEST_COMMAND_SET) {
            fprintf(stderr, KVM_MSG_UNKNOWN_DEBUGGER_COMMAND_SET);
            ERROR_THROW(0);
        }
        f2Array = (void **)funcArray[cmd->cmdSet];
    }
    if (cmd->cmd > (int)f2Array[0]) {
        fprintf(stderr, KVM_MSG_UNKNOWN_JDWP_COMMAND);
        ERROR_THROW(0);
    }
    func = (CommandHandler)(f2Array[cmd->cmd]);

    outStream_initReply(out, inStream_id(in));
    inStream_init(in);

    waitOnSuspend = FALSE;
        
    func(in, out);

    if (inStream_error(in)) {
        outStream_setError(out, inStream_error(in));
    }
    outStream_sendReply(out);
    inStream_destroy(in);
    outStream_destroy(out);
getOut:
    END_TEMPORARY_ROOTS
    return;
}

#endif /* ENABLE_JAVA_DEBUGGER */

