/*
 * Copyright (c) 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * 
 */

/*=========================================================================
 * SYSTEM:    KVM
 * SUBSYSTEM: Input stream for debugger interface
 * FILE:      debuggerInputStream.c
 * OVERVIEW:  Presents a stream-based interface to data coming
 *            from the user-level debugger.
 * AUTHOR:    From JPDA sources/modified by Bill Pittore, Sun Labs
 *=======================================================================*/

/*=========================================================================
 * Include files
 *=======================================================================*/

#include <global.h>

/*=========================================================================
 * Functions
 *=======================================================================*/

#if ENABLE_JAVA_DEBUGGER

/*
 * TO DO: Support processing of replies through command input streams.
 */
void
inStream_init(PacketInputStream *stream)
{
    /*
     * Copying packet root to the stream
     */

    stream->error = JDWP_Error_NONE;
    stream->segment = stream->packet->type.cmd.data;
    stream->left = stream->segment->length;
    stream->segment->index = 0;
}

long 
inStream_id(PacketInputStream *stream)
{
    return stream->packet->type.cmd.id;
}

unsigned char
inStream_command(PacketInputStream *stream)
{
    return stream->packet->type.cmd.cmd;
}

static long 
readBytes(PacketInputStream *stream, void *dest, long size) 
{
    /*
     * Iteration handles items that span multiple packet segments
     */
    if (stream->error) {
        return stream->error;
    }
    while (size > 0) {
        long count = MIN(size, stream->left);
        if (count == 0) {
            /* end of input */
            stream->error = JDWP_Error_INTERNAL;
            return stream->error;
        }
        if (dest) {
            memcpy(dest, &stream->segment->data[stream->segment->index], count);
        }
        stream->segment->index += count;
        stream->left -= count;
        if (stream->left == 0) {
            /*
             * Move to the next segment
             */
            stream->segment = stream->segment->next;
            if (stream->segment) {
                stream->left = stream->segment->length;
        stream->segment->index = 0;
            }
        }
        size -= count;
        if (dest) {
            dest = (char *)dest + count;
        }
    }
    return stream->error;
}

int
inStream_skipBytes(PacketInputStream *stream, int size) {
    return readBytes(stream, NULL, size);
}

bool_t 
inStream_readBoolean(PacketInputStream *stream)
{
    unsigned char flag;
    readBytes(stream, &flag, sizeof(flag));
    if (stream->error) {
        return 0;
    } else {
        return flag ? TRUE : FALSE;
    }
}

unsigned char 
inStream_readByte(PacketInputStream *stream)
{
    unsigned char val = 0;
    readBytes(stream, &val, sizeof(val));
    return val;
}

char 
inStream_readChar(PacketInputStream *stream)
{
    char val = 0;
    readBytes(stream, &val, sizeof(val));
    return JAVA_TO_HOST_CHAR(val);
}

short 
inStream_readShort(PacketInputStream *stream)
{
    short val = 0;
    readBytes(stream, &val, sizeof(val));
    return JAVA_TO_HOST_SHORT(val);
}

int 
inStream_readInt(PacketInputStream *stream)
{
    int val = 0;
    readBytes(stream, &val, sizeof(val));
    return JAVA_TO_HOST_INT(val);
}

long
inStream_readLong(PacketInputStream *stream)
{
    long val = 0;
    readBytes(stream, &val, sizeof(val));
    return JAVA_TO_HOST_INT(val);
}

/*  val1 is lower 32 bits, val2 is upper 32 bits */
void
inStream_readLong64(PacketInputStream *stream, unsigned long *val1, unsigned long *val2)
{

    readBytes(stream, val2, sizeof(val2));
    readBytes(stream, val1, sizeof(val1));
    *val1 = JAVA_TO_HOST_INT(*val1);
    *val2 = JAVA_TO_HOST_INT(*val2);
}

/*
 * Read an object from the stream. The ID used in the wire protocol
 * is converted to a reference which is returned. The reference is 
 * global and strong, but it should *not* be deleted by the caller
 * since it is freed when this stream is destroyed. 
 */
OBJECT 
inStream_readObjectRef(PacketInputStream *stream)
{
    long id = inStream_readLong(stream);
    if (stream->error) {
        return NULL;
    } 
    if (id == 0) {
        return NULL;
    }
    return (OBJECT)id;
}

/*
 * Read a raw object id from the stream. This should be used rarely.
 * Normally, inStream_readObjectRef is preferred since it takes care
 * of reference conversion and tracking. Only code that needs to 
 * perform maintenance of the commonRef hash table uses this function.
 */
long 
inStream_readObjectID(PacketInputStream *stream)
{
    return inStream_readLong(stream);
}

CLASS 
inStream_readClassRef(PacketInputStream *stream)
{
    OBJECT object = inStream_readObjectRef(stream);
    if (object == NULL) {
        /* 
         * Could be error or just the null reference. In either case,
         * stop now.
         */
        return NULL;
    }
    if (!isAssignableTo(object->ofClass, (CLASS)JavaLangClass)) {
        stream->error = JDWP_Error_INVALID_CLASS;
        return NULL;
    }
    return (CLASS)object;
}

THREAD 
inStream_readThreadRef(PacketInputStream *stream)
{
    OBJECT object = inStream_readObjectRef(stream);
    if (object == NULL) {
        /* 
         * Could be error or just the null reference. In either case,
         * stop now.
         */
        return NULL;
    }
  /*  if (!isThread(object)) {
       stream->error = JVMDI_ERROR_INVALID_THREAD;
        return NULL;
    } */
    return (THREAD)object;
}

#if 0
jthreadGroup 
inStream_readThreadGroupRef(PacketInputStream *stream)
{
    OBJECT object = inStream_readObjectRef(stream);
    if (object == NULL) {
        /* 
         * Could be error or just the null reference. In either case,
         * stop now.
         */
        return NULL;
    }
    if (!isThreadGroup(object)) {
        stream->error = JVMDI_ERROR_INVALID_THREAD_GROUP;
        return NULL;
    }
    return object;
}
#endif

STRING_INSTANCE
inStream_readStringRef(PacketInputStream *stream)
{
    OBJECT object = inStream_readObjectRef(stream);
    if (object == NULL) {
        /* 
         * Could be error or just the null reference. In either case,
         * stop now.
         */
        return NULL;
    }
    if (!isAssignableTo(object->ofClass, (CLASS)JavaLangString)) {
        stream->error = JDWP_Error_INVALID_STRING;
        return NULL;
    }
    return (STRING_INSTANCE)object;
}

#if 0
CLASS 
inStream_readClassLoaderRef(PacketInputStream *stream)
{
    OBJECT object = inStream_readObjectRef(stream);
    if (object == NULL) {
        /* 
         * Could be error or just the null reference. In either case,
         * stop now.
         */
        return NULL;
    }
    if (!isClassLoader(object)) {
        stream->error = JDWP_Error_INVALID_CLASS_LOADER;
        return NULL;
    }
    return object;
}
#endif

ARRAY 
inStream_readArrayRef(PacketInputStream *stream)
{
    OBJECT object = inStream_readObjectRef(stream);
    if (object == NULL) {
        /* 
         * Could be error or just the null reference. In either case,
         * stop now.
         */
        return NULL;
    }
    if (!IS_ARRAY_CLASS(object)) {
        stream->error = JDWP_Error_INVALID_ARRAY;
        return NULL;
    }
    return (ARRAY)object;
}

FRAME 
inStream_readFrameID(PacketInputStream *stream)
{
    return (FRAME)inStream_readLong(stream);
}

METHOD 
inStream_readMethodID(PacketInputStream *stream)
{
    return (METHOD)inStream_readLong(stream);
}

FIELD 
inStream_readFieldID(PacketInputStream *stream)
{
    return (FIELD)inStream_readLong(stream);
}

long 
inStream_readLocation(PacketInputStream *stream)
{
    return inStream_readLong(stream);
}

char * 
inStream_readString(PacketInputStream *stream)
{
    long length;
    char *string;

    START_TEMPORARY_ROOTS
    length = inStream_readLong(stream);
    IS_TEMPORARY_ROOT(stream, stream);
    IS_TEMPORARY_ROOT(string, mallocBytes(length + 1));
    if (string != NULL) {
        readBytes(stream, string, length);
        string[length] = '\0';
    }
    END_TEMPORARY_ROOTS
    return string;
}

bool_t 
inStream_endOfInput(PacketInputStream *stream)
{
    return (stream->left > 0);
}

int 
inStream_error(PacketInputStream *stream)
{
    return stream->error;
}

Value 
inStream_readValue(PacketInputStream *stream, unsigned char *typeKeyPtr)
{
    Value value;
    unsigned char typeKey = inStream_readByte(stream);
    if (stream->error) {
        value.u.j1 = 0;
        value.u.j2 = 0;
        return value;
    }

    if (isObjectTag(typeKey)) {
        value.l = inStream_readObjectRef(stream);
    } else {
        switch (typeKey) {
            case JDWP_Tag_BYTE:
                value.b = inStream_readByte(stream);
                break;
    
            case JDWP_Tag_CHAR:
                value.c = inStream_readChar(stream);
                break;
    
            case JDWP_Tag_INT:
                value.i = inStream_readLong(stream);
                break;
    
            case JDWP_Tag_LONG:
#if BIG_ENDIAN
                value.u.j1 = inStream_readLong(stream);
                value.u.j2 = inStream_readLong(stream);
#else
                value.u.j2 = inStream_readLong(stream);
                value.u.j1 = inStream_readLong(stream);
#endif
                break;
    
            case JDWP_Tag_SHORT:
                value.s = inStream_readShort(stream);
                break;
    
            case JDWP_Tag_BOOLEAN:
                value.z = inStream_readBoolean(stream);
                break;
            default:
                stream->error = JDWP_Error_INVALID_TAG;
                break;
        }
    }
    if (typeKeyPtr) {
        *typeKeyPtr = typeKey;
    }
    return value;
}

void 
inStream_destroy(PacketInputStream *stream)
{
    PacketData **segment;
    PacketData **nextSegment;

    segment = &stream->segment;
    while (*segment) {
        nextSegment = &(*segment)->next;
        *segment = NULL;
        segment = nextSegment;
    }
}

#endif /* ENABLE_JAVA_DEBUGGER */

