/*
 * Copyright (c) 1997-2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * 
 */

/*=========================================================================
 * SYSTEM:    KVM
 * SUBSYSTEM: Output stream for debugger interface
 * FILE:      debuggerOutputStream.c
 * OVERVIEW:  Presents a stream-based interface to data going 
 *            to the user level debugger.
 * AUTHOR:    From JPDA sources/modified by Bill Pittore, Sun Labs
 *=======================================================================*/

/*=========================================================================
 * Include files
 *=======================================================================*/

#include <global.h>

/*=========================================================================
 * Functions
 *=======================================================================*/

#if ENABLE_JAVA_DEBUGGER

#define INITIAL_ID_ALLOC  50

static void
commonInit(PacketOutputStream *stream)
{
    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(stream, stream);
    stream->left = INITIAL_SEGMENT_SIZE;
    stream->packet = (Packet *)callocObject(SIZEOF_PACKET, GCT_POINTERLIST);
    stream->numPointers = 1;
    stream->segment = stream->packet->type.cmd.data =
        (PacketData *)callocObject(SIZEOF_PACKETDATA, GCT_POINTERLIST);
    stream->numPointers = 2;
    stream->packet->type.cmd.numPointers = 1;
    stream->segment->length = stream->segment->index = 0;
    stream->segment->data = (unsigned char *)mallocBytes(INITIAL_SEGMENT_SIZE);
    stream->segment->numPointers = 1;
    stream->error = JDWP_Error_NONE;
    END_TEMPORARY_ROOTS
    return;
}

void
outStream_initCommand(PacketOutputStream *stream, long id, 
       unsigned char flags, unsigned char commandSet, unsigned char command)
{

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(stream, stream);
    commonInit(stream);

    /*
     * Command-specific initialization
     */
    stream->packet->type.cmd.id = id;
    stream->packet->type.cmd.cmdSet = commandSet;
    stream->packet->type.cmd.cmd = command;

    stream->packet->type.cmd.flags = flags;
    END_TEMPORARY_ROOTS
}

void 
outStream_initReply(PacketOutputStream *stream, long id)
{

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(stream, stream);
    commonInit(stream);

    /*
     * Reply-specific initialization
     */
    stream->packet->type.reply.id = id;
    stream->packet->type.reply.errorCode = REPLY_NoError;
    stream->packet->type.cmd.flags = FLAGS_Reply;
    END_TEMPORARY_ROOTS
}

long 
outStream_id(PacketOutputStream *stream)
{
    return stream->packet->type.cmd.id;
}

unsigned char 
outStream_command(PacketOutputStream *stream)
{
    /* Only makes sense for commands */
    return stream->packet->type.cmd.cmd;
}

static int 
writeBytes(PacketOutputStream *stream, void *source, long size, bool_t isHandle)
{
    unsigned char *bytes;
    unsigned long index = 0;
    unsigned char *newSeg;

    if (stream->error) {
        return stream->error;
    }
    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(stream, stream);
    if (isHandle)
        bytes = (unsigned char *)unhand((CONST_CHAR_HANDLE)source);
    else
        bytes = (unsigned char *)source;
    while (size > 0) {
        long count;
        if (stream->left == 0) {
            int segSize = MAX_SEGMENT_SIZE;
            stream->segment->next =
                (struct PacketData *)callocObject(SIZEOF_PACKETDATA,
                GCT_POINTERLIST);
            stream->segment->numPointers = 2;
            newSeg = (unsigned char *)mallocBytes(segSize);
            stream->segment->next->data = newSeg;
            stream->segment = stream->segment->next;
            stream->segment->numPointers = 1;
            stream->left = segSize;
            if (isHandle)
                bytes = (unsigned char *)unhand((CONST_CHAR_HANDLE)source);
            }
        count = MIN(size, stream->left);
        memcpy(&stream->segment->data[stream->segment->index], &bytes[index],
            count);
        stream->segment->index += count;
        stream->left -= count;
        stream->segment->length += count;
        size -= count;
        index += count;
    }
    END_TEMPORARY_ROOTS
    return (stream->error);
}

int 
outStream_writeBoolean(PacketOutputStream *stream, bool_t val)
{
    unsigned char byte = (val != 0) ? 1 : 0;
    return writeBytes(stream, &byte, sizeof(byte), FALSE);
}

int 
outStream_writeByte(PacketOutputStream *stream, unsigned char val)
{
    return writeBytes(stream, &val, sizeof(val), FALSE);
}

int 
outStream_writeChar(PacketOutputStream *stream, short val)
{
    val = HOST_TO_JAVA_CHAR(val);
    return writeBytes(stream, &val, sizeof(val), FALSE);
}

int 
outStream_writeShort(PacketOutputStream *stream, short val)
{
    val = HOST_TO_JAVA_SHORT(val);
    return writeBytes(stream, &val, sizeof(val), FALSE);
}

int 
outStream_writeInt(PacketOutputStream *stream, int val)
{
    val = HOST_TO_JAVA_INT(val);
    return writeBytes(stream, &val, sizeof(val), FALSE);
}

int 
outStream_writeLong(PacketOutputStream *stream, long val)
{
    val = HOST_TO_JAVA_INT(val);
    return writeBytes(stream, &val, sizeof(val), FALSE);
}

/* only called to write the code offset which is only 32 bits, so we fake the
 * upper word
 */
int 
outStream_writeLong64(PacketOutputStream *stream, unsigned long val)
{
    int retval = 0;
    long zero = 0;

    val = HOST_TO_JAVA_INT(val);
    retval = writeBytes(stream, &zero, sizeof(zero), FALSE);
    retval += writeBytes(stream, &val, sizeof(val), FALSE);
    return retval;
}

int 
outStream_writeObjectTag(PacketOutputStream *stream, OBJECT val)
{
    return outStream_writeByte(stream, specificTypeKey(val));
}

int 
outStream_writeObjectRef(PacketOutputStream *stream, OBJECT val)
{
    long id;

    if (stream->error) {
        return stream->error;
    }

    if (val == NULL) {
        id = 0;
    } else {
        /* Convert the object to an object id */
        /* Add the encoded object id to the stream */
        id = HOST_TO_JAVA_INT((long)val);
    }

    return writeBytes(stream, &id, sizeof(id), FALSE);
}

int 
outStream_writeClassID(PacketOutputStream *stream, CLASS val)
{
    return outStream_writeObjectRef(stream, (OBJECT)val);
}

int 
outStream_writeFrameID(PacketOutputStream *stream, FRAME val)
{
    return outStream_writeLong(stream, (long)val);
}

int 
outStream_writeMethodID(PacketOutputStream *stream, METHOD val)
{
    return outStream_writeLong(stream, (long)val);
}

int 
outStream_writeFieldID(PacketOutputStream *stream, FIELD val)
{
    return outStream_writeLong(stream, (long)val);
}

int 
outStream_writeLocation(PacketOutputStream *stream, 
    BYTE tag, CLASS clazz, METHOD method, unsigned long loc)
{
    int ret = 0;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(stream, stream);
    
    ret = outStream_writeByte(stream, tag);
    ret += outStream_writeClassID(stream, clazz);
    ret += outStream_writeMethodID(stream, method);
    ret += outStream_writeLong64(stream, loc);
    END_TEMPORARY_ROOTS
    return ret;
}

int 
outStream_writeByteArray(PacketOutputStream*stream, int length, 
                         unsigned char *bytes)
{
    int ret;
    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(stream, stream);
    outStream_writeInt(stream, length);
    ret = writeBytes(stream, bytes, length, FALSE);
    END_TEMPORARY_ROOTS
    return ret;
}

int 
outStream_writeString(PacketOutputStream *stream, CONST_CHAR_HANDLE string)
{
    int ret;
    long length = strlen((char *)unhand(string));
    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(stream, stream);
    outStream_writeLong(stream, length);
    ret = writeBytes(stream, (void **)string, length, TRUE);
    END_TEMPORARY_ROOTS
    return ret;
}

void
outStream_writeObjectValue(PacketOutputStream *stream, OBJECT object,
    unsigned char typeKey)
{
    unsigned int objectID;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(stream, stream);
    if (inCurrentHeap(object))
        IS_TEMPORARY_ROOT(object, object);
    if (validThread((THREAD)object))
        IS_TEMPORARY_ROOT(object, (OBJECT)((THREAD)(object))->javaThread);
    objectID = getObjectID(object);
    if (typeKey == JDWP_Tag_OBJECT)
        outStream_writeByte(stream, specificTypeKey(object));
    else
        outStream_writeByte(stream, typeKey);
    outStream_writeObjectRef(stream, (OBJECT)objectID);
    END_TEMPORARY_ROOTS

}

void 
outStream_writeValue(PacketOutputStream *stream, 
    unsigned char typeKey, void *value, void *value2, int writeTag )
{
    long v1, v2;

    /* pre-read values to avoid GC problems */

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(stream, stream);
    IS_TEMPORARY_ROOT(value2, value2);

    v1 = *(cell *)value;
    if (value2 != NULL)
        v2 = *(cell *)value2;

    if ( writeTag != 0 )
        outStream_writeByte(stream, typeKey);
    
    switch (typeKey) {
        case JDWP_Tag_BYTE:
            outStream_writeByte(stream, (unsigned char)(v1 & 0xFF));
            break;

        case JDWP_Tag_CHAR:
            outStream_writeChar(stream, (short)(v1 & 0xFFFF));
            break;

        case JDWP_Tag_INT:
        case JDWP_Tag_FLOAT:
            outStream_writeLong(stream, v1);
            break;
 
        case JDWP_Tag_LONG:
        case JDWP_Tag_DOUBLE:
#if BIG_ENDIAN
            outStream_writeLong( stream, v1 );
            outStream_writeLong( stream, v2 );
#else
            outStream_writeLong( stream, v2 );
            outStream_writeLong( stream, v1 );
#endif
            break;

        case JDWP_Tag_SHORT:
            outStream_writeShort(stream, (short)(v1 & 0xFFFF ));
            break;

        case JDWP_Tag_BOOLEAN:
            outStream_writeBoolean(stream,  v1 & 0xFF );
            break;

        case JDWP_Tag_VOID:  /* happens with function return values */   
            /* write nothing */
            break;

        default:
            fatalError(KVM_MSG_INVALID_JDWP_TYPE_TAG);
    }
    END_TEMPORARY_ROOTS
}

int 
outStream_skipBytes(PacketOutputStream *stream, int count)
{
    int i;

    START_TEMPORARY_ROOTS
    IS_TEMPORARY_ROOT(stream, stream);
    for (i = 0; i < count; i++) {
        outStream_writeByte(stream, 0);
    }
    END_TEMPORARY_ROOTS
    return stream->error;
}

int
outStream_error(PacketOutputStream *stream)
{
    return stream->error;
}

void 
outStream_setError(PacketOutputStream *stream, int error)
{
    if (!stream->error) {
        stream->error = error;
    }
}

void 
outStream_sendReply(PacketOutputStream *stream)
{
    int error;
    if (stream->error) {
        /*
         * Don't send any collected stream data on an error reply
         */
        stream->packet->type.reply.data->length = 0;
        stream->packet->type.reply.errorCode = (short)stream->error;
    } 
    error = dbgSendPacket(stream->packet);
}

void 
outStream_sendCommand(PacketOutputStream *stream)
{
    int error;
    if (!stream->error) {
        error = dbgSendPacket(stream->packet);
    } 
}

void 
outStream_destroy(PacketOutputStream *stream)
{
    PacketData **segment;
    PacketData **nextSegment;

    segment = &stream->segment;
    while (*segment) {
        nextSegment = &(*segment)->next;
        *segment = NULL;
        segment = nextSegment;
    }
}

long
uniqueID()
{
    static long currentID = 0;
    return currentID++;
}

#endif /* ENABLE_JAVA_DEBUGGER */

